"""

plot_utils.py

Utility functions for plotting, particularly for plotting confusion matrices
and precision-recall curves.

"""

#%% Imports

import numpy as np

# This also imports mpl.{cm, axes, colors}
import matplotlib.figure


#%% Plotting functions

def plot_confusion_matrix(matrix,
                          classes,
                          normalize=False,
                          title='Confusion matrix',
                          cmap=matplotlib.cm.Blues,
                          vmax=None,
                          use_colorbar=True,
                          y_label=True,
                          fmt= '{:.0f}',
                          fig=None):
    """
    Plots a confusion matrix.

    Args:
        matrix (np.ndarray): shape [num_classes, num_classes], confusion matrix
            where rows are ground-truth classes and columns are predicted classes
        classes (list of str): class names for each row/column
        normalize (bool, optional): whether to perform row-wise normalization;
            by default, assumes values in the confusion matrix are percentages
        title (str, optional): figure title
        cmap (matplotlib.colors.colormap, optional): colormap for cell backgrounds
        vmax (float, optional): value corresponding to the largest value of the colormap;
            if None, the maximum value in [matrix] will be used
        use_colorbar (bool, optional): whether to show colorbar
        y_label (bool, optional): whether to show class names on the y axis
        fmt (str, optional): format string for rendering numeric values
        fig (Figure, optional): existing figure to which we should render, otherwise
            creates a new figure

    Returns:
        matplotlib.figure.Figure: the figure we rendered to or created
    """

    num_classes = matrix.shape[0]
    assert matrix.shape[1] == num_classes
    assert len(classes) == num_classes

    normalized_matrix = matrix.astype(np.float64) / (
        matrix.sum(axis=1, keepdims=True) + 1e-7)
    if normalize:
        matrix = normalized_matrix

    fig_h = 3 + 0.3 * num_classes
    fig_w = fig_h
    if use_colorbar:
        fig_w += 0.5

    if fig is None:
        fig = matplotlib.figure.Figure(figsize=(fig_w, fig_h), tight_layout=True)
    ax = fig.subplots(1, 1)
    im = ax.imshow(normalized_matrix, interpolation='nearest', cmap=cmap, vmax=vmax)
    ax.set_title(title)

    if use_colorbar:
        cbar = fig.colorbar(im, fraction=0.046, pad=0.04,
                            ticks=[0.0, 0.25, 0.5, 0.75, 1.0])
        cbar.set_ticklabels(['0%', '25%', '50%', '75%', '100%'])

    tick_marks = np.arange(num_classes)
    ax.set_xticks(tick_marks)
    ax.set_yticks(tick_marks)
    ax.set_xticklabels(classes, rotation=90)
    ax.set_xlabel('Predicted class')

    if y_label:
        ax.set_yticklabels(classes)
        ax.set_ylabel('Ground-truth class')

    for i, j in np.ndindex(matrix.shape):
        v = matrix[i, j]
        ax.text(j, i, fmt.format(v),
                horizontalalignment='center',
                verticalalignment='center',
                color='white' if normalized_matrix[i, j] > 0.5 else 'black')

    return fig

# ...def plot_confusion_matrix(...)


def plot_precision_recall_curve(precisions,
                                recalls,
                                title='Precision/recall curve',
                                xlim=(0.0,1.05),
                                ylim=(0.0,1.05)):
    """
    Plots a precision/recall curve given lists of (ordered) precision
    and recall values.

    Args:
        precisions (list of float): precision for corresponding recall values,
            should have same length as [recalls].
        recalls (list of float): recall for corresponding precision values,
            should have same length as [precisions].
        title (str, optional): plot title
        xlim (tuple, optional): x-axis limits as a length-2 tuple
        ylim (tuple, optional): y-axis limits as a length-2 tuple

    Returns:
        matplotlib.figure.Figure: the (new) figure
    """

    assert len(precisions) == len(recalls)

    fig = matplotlib.figure.Figure(tight_layout=True)
    ax = fig.subplots(1, 1)
    ax.step(recalls, precisions, color='b', alpha=0.2, where='post')
    ax.fill_between(recalls, precisions, alpha=0.2, color='b', step='post')

    ax.set_xlabel('Recall')
    ax.set_ylabel('Precision')
    ax.set_title(title)
    ax.set_xlim(xlim[0],xlim[1])
    ax.set_ylim(ylim[0],ylim[1])

    return fig

# ...def plot_precision_recall_curve(...)


def plot_stacked_bar_chart(data,
                           series_labels=None,
                           col_labels=None,
                           x_label=None,
                           y_label=None,
                           log_scale=False):
    """
    Plot a stacked bar chart, for plotting e.g. species distribution across locations.

    Reference: https://stackoverflow.com/q/44309507

    Args:
        data (np.ndarray or list of list): data to plot; rows (series) are species, columns
            are locations
        series_labels (list of str, optional): series labels, typically species names
        col_labels (list of str, optional): column labels, typically location names
        x_label (str, optional): x-axis label
        y_label (str, optional): y-axis label
        log_scale (bool, optional): whether to plot the y axis in log-scale

    Returns:
        matplotlib.figure.Figure: the (new) figure
    """

    data = np.asarray(data)
    num_series, num_columns = data.shape
    ind = np.arange(num_columns)

    fig = matplotlib.figure.Figure(tight_layout=True)
    ax = fig.subplots(1, 1)
    colors = matplotlib.cm.rainbow(np.linspace(0, 1, num_series))

    # stacked bar charts are made with each segment starting from a y position
    cumulative_size = np.zeros(num_columns)
    for i_row, row_data in enumerate(data):
        if series_labels is None:
            label = 'series_{}'.format(str(i_row).zfill(2))
        else:
            label = series_labels[i_row]
        ax.bar(ind, row_data, bottom=cumulative_size, label=label,
               color=colors[i_row])
        cumulative_size += row_data

    if (col_labels is not None) and (len(col_labels) < 25):
        ax.set_xticks(ind)
        ax.set_xticklabels(col_labels, rotation=90)
    elif (col_labels is not None):
        ax.set_xticks(list(range(0, len(col_labels), 20)))
        ax.set_xticklabels(col_labels[::20], rotation=90)

    if x_label is not None:
        ax.set_xlabel(x_label)
    if y_label is not None:
        ax.set_ylabel(y_label)
    if log_scale:
        ax.set_yscale('log')

    # To fit the legend in, shrink current axis by 20%
    box = ax.get_position()
    ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])

    # Put a legend to the right of the current axis
    ax.legend(loc='center left', bbox_to_anchor=(0.99, 0.5), frameon=False)

    return fig

# ...def plot_stacked_bar_chart(...)


def calibration_ece(true_scores, pred_scores, num_bins):
    r"""
    Expected calibration error (ECE) as defined in equation (3) of
    Guo et al. "On Calibration of Modern Neural Networks." (2017).

    Implementation modified from sklearn.calibration.calibration_curve()
    in order to implement ECE calculation. See:

    https://github.com/scikit-learn/scikit-learn/issues/18268

    Args:
        true_scores (list of int): true values, length N, binary-valued (0 = neg, 1 = pos)
        pred_scores (list of float): predicted confidence values, length N, pred_scores[i] is the
            predicted confidence that example i is positive
        num_bins (int): number of bins to use (`M` in eq. (3) of Guo 2017)

    Returns:
        tuple: a length-three tuple containing:
            - accs: np.ndarray, shape [M], type float64, accuracy in each bin,
              M <= num_bins because bins with no samples are not returned
            - confs: np.ndarray, shape [M], type float64, mean model confidence in
              each bin
            - ece: float, expected calibration error
    """

    assert len(true_scores) == len(pred_scores)

    bins = np.linspace(0., 1. + 1e-8, num=num_bins + 1)
    binids = np.digitize(pred_scores, bins) - 1

    bin_sums = np.bincount(binids, weights=pred_scores, minlength=len(bins))
    bin_true = np.bincount(binids, weights=true_scores, minlength=len(bins))
    bin_total = np.bincount(binids, minlength=len(bins))

    nonzero = bin_total != 0
    accs = bin_true[nonzero] / bin_total[nonzero]
    confs = bin_sums[nonzero] / bin_total[nonzero]

    weights = bin_total[nonzero] / len(true_scores)
    ece = np.abs(accs - confs) @ weights
    return accs, confs, ece

# ...def calibration_ece(...)


def plot_calibration_curve(true_scores,
                           pred_scores,
                           num_bins,
                           name='calibration',
                           plot_perf=True,
                           plot_hist=True,
                           ax=None,
                           **fig_kwargs):
    """
    Plots a calibration curve.

    Args:
        true_scores (list of int): true values, length N, binary-valued (0 = neg, 1 = pos)
        pred_scores (list of float): predicted confidence values, length N, pred_scores[i] is the
            predicted confidence that example i is positive
        num_bins (int): number of bins to use (`M` in eq. (3) of Guo 2017)
        name (str, optional): label in legend for the calibration curve
        plot_perf (bool, optional): whether to plot y=x line indicating perfect calibration
        plot_hist (bool, optional): whether to plot histogram of counts
        ax (Axes, optional): if given then no legend is drawn, and fig_kwargs are ignored
        fig_kwargs (dict): only used if [ax] is None

    Returns:
        matplotlib.figure.Figure: the (new) figure
    """

    accs, confs, ece = calibration_ece(true_scores, pred_scores, num_bins)

    created_fig = False
    if ax is None:
        created_fig = True
        fig = matplotlib.figure.Figure(**fig_kwargs)
        ax = fig.subplots(1, 1)
    ax.plot(confs, accs, 's-', label=name)  # 's-': squares on line
    ax.set(xlabel='Model confidence', ylabel='Actual accuracy',
           title=f'Calibration plot (ECE: {ece:.02g})')
    ax.set(xlim=[-0.05, 1.05], ylim=[-0.05, 1.05])
    if plot_perf:
        ax.plot([0, 1], [0, 1], color='black', label='perfect calibration')
    ax.grid(True)

    if plot_hist:
        ax1 = ax.twinx()
        bins = np.linspace(0., 1. + 1e-8, num=num_bins + 1)
        counts = ax1.hist(pred_scores, alpha=0.5, label='histogram of examples',
                          bins=bins, color='tab:red')[0]
        max_count = np.max(counts)
        ax1.set_ylim([-0.05 * max_count, 1.05 * max_count])
        ax1.set_ylabel('Count')

    if created_fig:
        fig.legend(loc='upper left', bbox_to_anchor=(0.15, 0.85))

    return ax.figure

# ...def plot_calibration_curve(...)



"""

video_utils.py

Utilities for splitting, rendering, and assembling videos.

"""

#%% Constants, imports, environment

import os
import re
import cv2
import glob
import json

from collections import defaultdict
from multiprocessing.pool import ThreadPool
from multiprocessing.pool import Pool
from tqdm import tqdm
from functools import partial
from inspect import signature

from megadetector.utils import path_utils
from megadetector.utils.path_utils import clean_path
from megadetector.utils.ct_utils import sort_list_of_dicts_by_key
from megadetector.visualization import visualization_utils as vis_utils

default_fourcc = 'h264'

video_progress_bar_description = 'Processing video'

#%% Path utilities

VIDEO_EXTENSIONS = ('.mp4','.avi','.mpeg','.mpg','.mov','.mkv','.flv')

def is_video_file(s,video_extensions=VIDEO_EXTENSIONS):
    """
    Checks a file's extension against a set of known video file
    extensions to determine whether it's a video file.  Performs a
    case-insensitive comparison.

    Args:
        s (str): filename to check for probable video-ness
        video_extensions (list, optional): list of video file extensions

    Returns:
        bool: True if this looks like a video file, else False
    """

    ext = os.path.splitext(s)[1]
    return ext.lower() in video_extensions


def find_video_strings(strings):
    """
    Given a list of strings that are potentially video file names, looks for
    strings that actually look like video file names (based on extension).

    Args:
        strings (list): list of strings to check for video-ness

    Returns:
        list: a subset of [strings] that looks like they are video filenames
    """

    return [s for s in strings if is_video_file(s.lower())]


def find_videos(dirname,
                recursive=False,
                convert_slashes=True,
                return_relative_paths=False):
    """
    Finds all files in a directory that look like video file names.

    Args:
        dirname (str): folder to search for video files
        recursive (bool, optional): whether to search [dirname] recursively
        convert_slashes (bool, optional): forces forward slashes in the returned files,
            otherwise uses the native path separator
        return_relative_paths (bool, optional): forces the returned filenames to be
            relative to [dirname], otherwise returns absolute paths

    Returns:
        A list of filenames within [dirname] that appear to be videos
    """

    if recursive:
        files = glob.glob(os.path.join(dirname, '**', '*.*'), recursive=True)
    else:
        files = glob.glob(os.path.join(dirname, '*.*'))

    files = [fn for fn in files if os.path.isfile(fn)]

    if return_relative_paths:
        files = [os.path.relpath(fn,dirname) for fn in files]

    if convert_slashes:
        files = [fn.replace('\\', '/') for fn in files]

    return find_video_strings(files)


#%% Shared function for opening videos

DEFAULT_BACKEND = -1

# This is the order in which we'll try to open backends.
#
# In general, the defaults are as follows, though they vary depending
# on what's installed:
#
# Windows: CAP_DSHOW or CAP_MSMF
# Linux: CAP_FFMPEG
# macOS: CAP_AVFOUNDATION
#
# Technically if the default fails, we may try the same backend again, but this
# is rare, and it's not worth the complexity of figuring out what the system
# default is.
backend_id_to_name = {
    DEFAULT_BACKEND:'default',
    cv2.CAP_FFMPEG: 'CAP_FFMPEG',
    cv2.CAP_DSHOW: 'CAP_DSHOW',
    cv2.CAP_MSMF: 'CAP_MSMF',
    cv2.CAP_AVFOUNDATION: 'CAP_AVFOUNDATION',
    cv2.CAP_GSTREAMER: 'CAP_GSTREAMER'
}

def open_video(video_path,verbose=False):
    """
    Open the video at [video_path], trying multiple OpenCV backends if necessary.

    Args:
        video_path (str): the file to open
        verbose (bool, optional): enable additional debug output

    Returns:
        (cv2.VideoCapture,image): a tuple containing (a) the open video capture device
        (or None if no backends succeeded) and (b) the first frame of the video (or None)
    """

    if not os.path.isfile(video_path):
        print('Video file {} not found'.format(video_path))
        return None,None

    backend_ids = backend_id_to_name.keys()

    for backend_id in backend_ids:

        backend_name = backend_id_to_name[backend_id]
        if verbose:
            print('Trying backend {}'.format(backend_name))

        try:
            if backend_id == DEFAULT_BACKEND:
                vidcap = cv2.VideoCapture(video_path)
            else:
                vidcap = cv2.VideoCapture(video_path, backend_id)
        except Exception as e:
            if verbose:
                print('Warning: error opening {} with backend {}: {}'.format(
                    video_path,backend_name,str(e)))
            continue

        if not vidcap.isOpened():
            if verbose:
                print('Warning: isOpened() is False for {} with backend {}'.format(
                    video_path,backend_name))
            try:
                vidcap.release()
            except Exception:
                pass
            continue

        success, image = vidcap.read()
        if success and (image is not None):
            if verbose:
                print('Successfully opened {} with backend: {}'.format(
                    video_path,backend_name))
            return vidcap,image

        print('Warning: failed to open {} with backend {}'.format(
            video_path,backend_name))
        try:
            vidcap.release()
        except Exception:
            pass

    # ...for each backend

    print('Error: failed to open {} with any backend'.format(video_path))
    return None,None

# ...def open_video(...)


#%% Functions for rendering frames to video and vice-versa

# http://tsaith.github.io/combine-images-into-a-video-with-python-3-and-opencv-3.html

def frames_to_video(images, fs, output_file_name, codec_spec=default_fourcc):
    """
    Given a list of image files and a sample rate, concatenates those images into
    a video and writes to a new video file.

    Args:
        images (list): a list of frame file names to concatenate into a video
        fs (float): the frame rate in fps
        output_file_name (str): the output video file, no checking is performed to make
            sure the extension is compatible with the codec
        codec_spec (str, optional):  codec to use for encoding; h264 is a sensible default
            and generally works on Windows, but when this fails (which is around 50% of the time
            on Linux), mp4v is a good second choice
    """

    if codec_spec is None:
        codec_spec = 'h264'

    if len(images) == 0:
        print('Warning: no frames to render')
        return

    output_dir = os.path.dirname(output_file_name)
    if len(output_dir) > 0:
        os.makedirs(output_dir, exist_ok=True)

    # Determine the width and height from the first image
    frame = cv2.imread(images[0])
    # cv2.imshow('video',frame)
    height, width, channels = frame.shape

    # Define the codec and create VideoWriter object
    fourcc = cv2.VideoWriter_fourcc(*codec_spec)
    out = cv2.VideoWriter(output_file_name, fourcc, fs, (width, height))

    for image in images:
        frame = cv2.imread(image)
        out.write(frame)

    out.release()
    cv2.destroyAllWindows()


def get_video_fs(input_video_file,verbose=False):
    """
    Retrieves the frame rate of [input_video_file].

    Args:
        input_video_file (str): video file for which we want the frame rate
        verbose (bool, optional): enable additional debug output

    Returns:
        float: the frame rate of [input_video_file], or None if no frame
            rate could be extracted
    """

    assert os.path.isfile(input_video_file), \
        'File {} not found'.format(input_video_file)
    vidcap,_ = open_video(input_video_file,verbose=verbose)
    if vidcap is None:
        if verbose:
            print('Failed to get frame rate for {}'.format(input_video_file))
        return None
    fs = vidcap.get(cv2.CAP_PROP_FPS)
    try:
        vidcap.release()
    except Exception as e:
        print('Warning: error closing video handle for {}: {}'.format(
            input_video_file,str(e)))
    return fs


def _frame_number_to_filename(frame_number):
    """
    Ensures that frame images are given consistent filenames.
    """

    return 'frame{:06d}.jpg'.format(frame_number)


def _filename_to_frame_number(filename):
    """
    Extract the frame number from a filename that was created using
    _frame_number_to_filename.

    Args:
        filename (str): a filename created with _frame_number_to_filename.

    Returns:
        int: the frame number extracted from [filename]
    """

    filename = os.path.basename(filename)
    match = re.search(r'frame(\d+)\.jpg', filename)
    if match is None:
        raise ValueError('{} does not appear to be a frame file'.format(filename))
    frame_number = match.group(1)
    try:
        frame_number = int(frame_number)
    except Exception:
        raise ValueError('Filename {} does not contain a valid frame number'.format(filename))

    return frame_number


def _add_frame_numbers_to_results(results):
    """
    Given the 'images' list from a set of MD results that was generated on video frames,
    add a 'frame_number' field to each image, and return the list, sorted by frame number.
    Also modifies "results" in place.

    Args:
        results (list): list of image dicts
    """

    # This indicate that this was a failure for a single video
    if isinstance(results,dict):
        assert 'failure' in results
        return results

    # Add video-specific fields to the results
    for im in results:
        fn = im['file']
        frame_number = _filename_to_frame_number(fn)
        im['frame_number'] = frame_number

    results = sort_list_of_dicts_by_key(results,'frame_number')
    return results


def run_callback_on_frames(input_video_file,
                           frame_callback,
                           every_n_frames=None,
                           verbose=False,
                           frames_to_process=None,
                           allow_empty_videos=False):
    """
    Calls the function frame_callback(np.array,image_id) on all (or selected) frames in
    [input_video_file].

    Args:
        input_video_file (str): video file to process
        frame_callback (function): callback to run on frames, should take an np.array and a string and
            return a single value.  callback should expect two arguments: (1) a numpy array with image
            data, in the typical PIL image orientation/channel order, and (2) a string identifier
            for the frame, typically something like "frame0006.jpg" (even though it's not a JPEG
            image, this is just an identifier for the frame).
        every_n_frames (int or float, optional): sample every Nth frame starting from the first frame;
            if this is None or 1, every frame is processed.  If this is a negative value, it's
            interpreted as a sampling rate in seconds, which is rounded to the nearest frame sampling
            rate. Mutually exclusive with frames_to_process.
        verbose (bool, optional): enable additional debug console output
        frames_to_process (list of int, optional): process this specific set of frames;
            mutually exclusive with every_n_frames.  If all values are beyond the length
            of the video, no frames are extracted.  Can also be a single int, specifying
            a single frame number.
        allow_empty_videos (bool, optional): Just print a warning if a video appears to have no
            frames (by default, this raises an Exception).

    Returns:
        dict: dict with keys 'frame_filenames' (list), 'frame_rate' (float), 'results' (list).
        'frame_filenames' are synthetic filenames (e.g. frame000000.jpg). Elements in
        'results' are whatever is returned by the callback, typically dicts in the same format used in
        the 'images' array in the MD results format.  [frame_filenames] and [results] both have
        one element per processed frame.
    """

    assert os.path.isfile(input_video_file), 'File {} not found'.format(input_video_file)

    if isinstance(frames_to_process,int):
        frames_to_process = [frames_to_process]

    if (frames_to_process is not None) and (every_n_frames is not None):
        raise ValueError('frames_to_process and every_n_frames are mutually exclusive')

    vidcap = None

    try:

        vidcap,image = open_video(input_video_file,verbose=verbose)
        n_frames = int(vidcap.get(cv2.CAP_PROP_FRAME_COUNT))
        frame_rate = vidcap.get(cv2.CAP_PROP_FPS)

        if verbose:
            print('Video {} contains {} frames at {} Hz'.format(input_video_file,n_frames,frame_rate))

        frame_filenames = []
        results = []

        if (every_n_frames is not None):

            if (every_n_frames < 0):
                every_n_seconds = abs(every_n_frames)
                every_n_frames = int(every_n_seconds * frame_rate)
                if verbose:
                    print('Interpreting a time sampling rate of {} hz as a frame interval of {}'.format(
                        every_n_seconds,every_n_frames))
            # 0 and 1 both mean "process every frame"
            elif every_n_frames == 0:
                every_n_frames = 1
            elif every_n_frames > 0:
                every_n_frames = int(every_n_frames)

        # ...if every_n_frames was supplied

        # frame_number = 0
        for frame_number in range(0,n_frames):

            # We've already read the first frame, when we opened the video
            if frame_number != 0:
                success,image = vidcap.read()
            else:
                success = True

            if not success:
                assert image is None
                if verbose:
                    print('Read terminating at frame {} of {}'.format(frame_number,n_frames))
                break

            if every_n_frames is not None:
                if (frame_number % every_n_frames) != 0:
                    continue

            if frames_to_process is not None:
                if frame_number > max(frames_to_process):
                    break
                if frame_number not in frames_to_process:
                    continue

            frame_filename_relative = _frame_number_to_filename(frame_number)
            frame_filenames.append(frame_filename_relative)

            # Convert from OpenCV conventions to PIL conventions
            image_np = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

            # Run the callback
            frame_results = frame_callback(image_np,frame_filename_relative)

            results.append(frame_results)

        # ...for each frame

        if len(frame_filenames) == 0:
            if allow_empty_videos:
                print('Warning: found no frames in file {}'.format(input_video_file))
            else:
                raise Exception('Error: found no frames in file {}'.format(input_video_file))

        if verbose:
            print('\nProcessed {} of {} frames for {}'.format(
                len(frame_filenames),n_frames,input_video_file))

    finally:

        if vidcap is not None:
            try:
                vidcap.release()
            except Exception:
                pass

    to_return = {}
    to_return['frame_filenames'] = frame_filenames
    to_return['frame_rate'] = frame_rate
    to_return['results'] = results

    return to_return

# ...def run_callback_on_frames(...)


def run_callback_on_frames_for_folder(input_video_folder,
                                      frame_callback,
                                      every_n_frames=None,
                                      verbose=False,
                                      recursive=True,
                                      files_to_process_relative=None,
                                      error_on_empty_video=False):
    """
    Calls the function frame_callback(np.array,image_id) on all (or selected) frames in
    all videos in [input_video_folder].

    Args:
        input_video_folder (str): video folder to process
        frame_callback (function): callback to run on frames, should take an np.array and a string and
            return a single value.  callback should expect two arguments: (1) a numpy array with image
            data, in the typical PIL image orientation/channel order, and (2) a string identifier
            for the frame, typically something like "frame0006.jpg" (even though it's not a JPEG
            image, this is just an identifier for the frame).
        every_n_frames (int or float, optional): sample every Nth frame starting from the first frame;
            if this is None or 1, every frame is processed.  If this is a negative value, it's
            interpreted as a sampling rate in seconds, which is rounded to the nearest frame
            sampling rate.
        verbose (bool, optional): enable additional debug console output
        recursive (bool, optional): recurse into [input_video_folder]
        files_to_process_relative (list, optional): only process specific relative paths
        error_on_empty_video (bool, optional): by default, videos with errors or no valid frames
            are silently stored as failures; this turns them into exceptions

    Returns:
        dict: dict with keys 'video_filenames' (list of str), 'frame_rates' (list of floats),
        'results' (list of list of dicts). 'video_filenames' will contain *relative* filenames.
        'results' is a list (one element per video) of lists (one element per frame) of whatever the
        callback returns, typically (but not necessarily) dicts in the MD results format.

        For failed videos, the frame rate will be represented by -1, and "results"
        will be a dict with at least the key "failure".
    """

    to_return = {'video_filenames':[],'frame_rates':[],'results':[]}

    if files_to_process_relative is not None:
        input_files_full_paths = \
            [os.path.join(input_video_folder,fn) for fn in files_to_process_relative]
        input_files_full_paths = [fn.replace('\\','/') for fn in input_files_full_paths]
    else:
        # Recursively enumerate video files
        input_files_full_paths = find_videos(input_video_folder,
                                            recursive=recursive,
                                            convert_slashes=True,
                                            return_relative_paths=False)

    print('Processing {} videos from folder {}'.format(len(input_files_full_paths),input_video_folder))

    if len(input_files_full_paths) == 0:
        print('No videos to process')
        return to_return

    # Process each video

    # video_fn_abs = input_files_full_paths[0]
    for video_fn_abs in tqdm(input_files_full_paths,desc=video_progress_bar_description):

        video_filename_relative = os.path.relpath(video_fn_abs,input_video_folder)
        video_filename_relative = video_filename_relative.replace('\\','/')
        to_return['video_filenames'].append(video_filename_relative)

        try:

            # video_results is a dict with fields:
            #
            # frame_rate
            #
            # results (list of objects returned by the callback, typically dicts in the MD
            # per-image format)
            #
            # frame_filenames (list of frame IDs, i.e. synthetic filenames)
            video_results = run_callback_on_frames(input_video_file=video_fn_abs,
                                                   frame_callback=frame_callback,
                                                   every_n_frames=every_n_frames,
                                                   verbose=verbose,
                                                   frames_to_process=None,
                                                   allow_empty_videos=False)

        except Exception as e:

            if (not error_on_empty_video):
                print('Warning: error processing video {}: {}'.format(
                    video_fn_abs,str(e)
                ))
                to_return['frame_rates'].append(-1.0)
                failure_result = {}
                failure_result['failure'] = 'Failure processing video: {}'.format(str(e))
                to_return['results'].append(failure_result)
                continue
            else:
                raise

        # ...try/except

        to_return['frame_rates'].append(video_results['frame_rate'])
        for r in video_results['results']:
            assert r['file'].startswith('frame')
            r['file'] = video_filename_relative + '/' + r['file']
        to_return['results'].append(video_results['results'])

    # ...for each video

    n_videos = len(input_files_full_paths)
    assert len(to_return['video_filenames']) == n_videos
    assert len(to_return['frame_rates']) == n_videos
    assert len(to_return['results']) == n_videos

    return to_return

# ...def run_callback_on_frames_for_folder(...)


def video_to_frames(input_video_file,
                    output_folder,
                    overwrite=True,
                    every_n_frames=None,
                    verbose=False,
                    quality=None,
                    max_width=None,
                    frames_to_extract=None,
                    allow_empty_videos=True):
    """
    Renders frames from [input_video_file] to .jpg files in [output_folder].

    With help from:

    https://stackoverflow.com/questions/33311153/python-extracting-and-saving-video-frames

    Args:
        input_video_file (str): video file to split into frames
        output_folder (str): folder to put frame images in
        overwrite (bool, optional): whether to overwrite existing frame images
        every_n_frames (int, optional): sample every Nth frame starting from the first frame;
            if this is None or 1, every frame is extracted.  If this is a negative value, it's
            interpreted as a sampling rate in seconds, which is rounded to the nearest frame sampling
            rate.  Mutually exclusive with frames_to_extract.
        verbose (bool, optional): enable additional debug console output
        quality (int, optional): JPEG quality for frame output, from 0-100.  Defaults
            to the opencv default (typically 95).
        max_width (int, optional): resize frames to be no wider than [max_width]
        frames_to_extract (list of int, optional): extract this specific set of frames;
            mutually exclusive with every_n_frames.  If all values are beyond the length
            of the video, no frames are extracted.  Can also be a single int, specifying
            a single frame number.  In the special case where frames_to_extract
            is [], this function still reads video frame rates and verifies that videos
            are readable, but no frames are extracted.
        allow_empty_videos (bool, optional): Just print a warning if a video appears to have
            no frames (by default, this is an error).

    Returns:
        tuple: length-2 tuple containing (list of frame filenames,frame rate)
    """

    assert os.path.isfile(input_video_file), 'File {} not found'.format(input_video_file)

    if quality is not None and quality < 0:
        quality = None

    if isinstance(frames_to_extract,int):
        frames_to_extract = [frames_to_extract]

    if (frames_to_extract is not None) and (every_n_frames is not None):
        raise ValueError('frames_to_extract and every_n_frames are mutually exclusive')

    bypass_extraction = ((frames_to_extract is not None) and (len(frames_to_extract) == 0))

    if not bypass_extraction:
        os.makedirs(output_folder,exist_ok=True)

    vidcap = cv2.VideoCapture(input_video_file)
    n_frames = int(vidcap.get(cv2.CAP_PROP_FRAME_COUNT))
    fs = vidcap.get(cv2.CAP_PROP_FPS)

    if (every_n_frames is not None) and (every_n_frames < 0):
        every_n_seconds = abs(every_n_frames)
        every_n_frames = int(every_n_seconds * fs)
        if verbose:
            print('Interpreting a time sampling rate of {} hz as a frame interval of {}'.format(
                every_n_seconds,every_n_frames))

    # If we're not over-writing, check whether all frame images already exist
    if (not overwrite) and (not bypass_extraction):

        missing_frame_number = None
        missing_frame_filename = None
        frame_filenames = []
        found_existing_frame = False

        for frame_number in range(0,n_frames):

            if every_n_frames is not None:
                assert frames_to_extract is None, \
                    'Internal error: frames_to_extract and every_n_frames are exclusive'
                if (frame_number % every_n_frames) != 0:
                    continue

            if frames_to_extract is not None:
                assert every_n_frames is None, \
                    'Internal error: frames_to_extract and every_n_frames are exclusive'
                if frame_number not in frames_to_extract:
                    continue

            frame_filename = _frame_number_to_filename(frame_number)
            frame_filename = os.path.join(output_folder,frame_filename)
            frame_filenames.append(frame_filename)
            if os.path.isfile(frame_filename):
                found_existing_frame = True
                continue
            else:
                missing_frame_number = frame_number
                missing_frame_filename = frame_filename
                break

        if verbose and missing_frame_number is not None:
            print('Missing frame {} ({}) for video {}'.format(
                missing_frame_number,
                missing_frame_filename,
                input_video_file))

        # OpenCV seems to over-report the number of frames by 1 in some cases, or fails
        # to read the last frame; either way, I'm allowing one missing frame.
        allow_last_frame_missing = True

        # This doesn't have to mean literally the last frame number, it just means that if
        # we find this frame or later, we consider the video done
        last_expected_frame_number = n_frames-1
        if every_n_frames is not None:
            last_expected_frame_number -= (every_n_frames*2)

        # When specific frames are requested, if anything is missing, reprocess the video
        if (frames_to_extract is not None) and (missing_frame_number is not None):
            pass

        # If no frames are missing, or only frames very close to the end of the video are "missing",
        # skip this video
        elif (missing_frame_number is None) or \
            (allow_last_frame_missing and (missing_frame_number >= last_expected_frame_number)):

            if verbose:
                print('Skipping video {}, all output frames exist'.format(input_video_file))
            return frame_filenames,fs

        else:

            # If we found some frames, but not all, print a message
            if verbose and found_existing_frame:
                print("Rendering video {}, couldn't find frame {} ({}) of {}".format(
                    input_video_file,
                    missing_frame_number,
                    missing_frame_filename,
                    last_expected_frame_number))

    # ...if we need to check whether to skip this video entirely

    if verbose:
        print('Video {} contains {} frames at {} Hz'.format(input_video_file,n_frames,fs))

    frame_filenames = []

    # YOLOv5 does some totally bananas monkey-patching of opencv, which causes
    # problems  if we try to supply a third parameter to imwrite (to specify JPEG
    # quality).  Detect  this case, and ignore the quality  parameter if it looks
    # like imwrite has been messed with.
    #
    # See:
    #
    # https://github.com/ultralytics/yolov5/issues/7285
    imwrite_patched = False
    n_imwrite_parameters = None

    try:
        # calling signature() on the native cv2.imwrite function will
        # fail, so an exception here is a good thing.  In fact I don't think
        # there's a case where this *succeeds* and the number of parameters
        # is wrong.
        sig = signature(cv2.imwrite)
        n_imwrite_parameters = len(sig.parameters)
    except Exception:
        pass

    if (n_imwrite_parameters is not None) and (n_imwrite_parameters < 3):
        imwrite_patched = True
        if verbose and (quality is not None):
            print('Warning: quality value supplied, but YOLOv5 has mucked with cv2.imwrite, ignoring quality')

    # for frame_number in tqdm(range(0,n_frames)):
    for frame_number in range(0,n_frames):

        # Special handling for the case where we're just doing dummy reads
        if bypass_extraction:
            break

        success,image = vidcap.read()
        if not success:
            assert image is None
            if verbose:
                print('Read terminating at frame {} of {}'.format(frame_number,n_frames))
            break

        if every_n_frames is not None:
            if (frame_number % every_n_frames) != 0:
                continue

        if frames_to_extract is not None:
            if frame_number > max(frames_to_extract):
                break
            if frame_number not in frames_to_extract:
                continue

        # Has resizing been requested?
        if max_width is not None:

            # image.shape is h/w/dims
            input_shape = image.shape
            assert input_shape[2] == 3
            input_width = input_shape[1]

            # Is resizing necessary?
            if input_width > max_width:

                scale = max_width / input_width
                assert scale <= 1.0

                # INTER_AREA is recommended for size reduction
                image = cv2.resize(image, (0,0), fx=scale, fy=scale, interpolation=cv2.INTER_AREA)

        # ...if we need to deal with resizing

        frame_filename_relative = _frame_number_to_filename(frame_number)
        frame_filename = os.path.join(output_folder,frame_filename_relative)
        frame_filenames.append(frame_filename)

        if (not overwrite) and (os.path.isfile(frame_filename)):
            # print('Skipping frame {}'.format(frame_filename))
            pass
        else:
            try:
                if frame_filename.isascii():

                    if quality is None or imwrite_patched:
                        cv2.imwrite(os.path.normpath(frame_filename),image)
                    else:
                        cv2.imwrite(os.path.normpath(frame_filename),image,
                                    [int(cv2.IMWRITE_JPEG_QUALITY), quality])
                else:
                    if quality is None:
                        is_success, im_buf_arr = cv2.imencode('.jpg', image)
                    else:
                        encode_param = [int(cv2.IMWRITE_JPEG_QUALITY), quality]
                        is_success, im_buf_arr = cv2.imencode('.jpg', image, encode_param)
                    im_buf_arr.tofile(frame_filename)
                assert os.path.isfile(frame_filename), \
                    'Output frame {} unavailable'.format(frame_filename)
            except KeyboardInterrupt:
                vidcap.release()
                raise
            except Exception as e:
                print('Error on frame {} of {}: {}'.format(frame_number,n_frames,str(e)))

    # ...for each frame

    if len(frame_filenames) == 0:
        if allow_empty_videos:
            print('Warning: no frames extracted from file {}'.format(input_video_file))
        else:
            raise Exception('Error: no frames extracted from file {}'.format(input_video_file))

    if verbose:
        print('\nExtracted {} of {} frames for {}'.format(
            len(frame_filenames),n_frames,input_video_file))

    vidcap.release()
    return frame_filenames,fs

# ...def video_to_frames(...)


def _video_to_frames_with_per_video_frames(args):
    """
    Wrapper function to handle extracting a different list of frames for
    each video in a multiprocessing context.

    Takes a tuple of (relative_fn, frames_for_this_video, other_args),
    where (other_args) contains the arguments that are the same for each
    iteration.
    """

    relative_fn, frames_for_this_video, other_args = args
    (input_folder, output_folder_base, every_n_frames, overwrite, verbose,
     quality, max_width, allow_empty_videos) = other_args

    return _video_to_frames_for_folder(relative_fn, input_folder, output_folder_base,
                                       every_n_frames, overwrite, verbose, quality, max_width,
                                       frames_for_this_video, allow_empty_videos)


def _video_to_frames_for_folder(relative_fn,input_folder,output_folder_base,
                                every_n_frames,overwrite,verbose,quality,max_width,
                                frames_to_extract,allow_empty_videos):
    """
    Internal function to call video_to_frames for a single video in the context of
    video_folder_to_frames;  makes sure the right output folder exists, then calls
    video_to_frames.
    """

    input_fn_absolute = os.path.join(input_folder,relative_fn)
    assert os.path.isfile(input_fn_absolute),\
        'Could not find file {}'.format(input_fn_absolute)

    # Create the target output folder
    output_folder_video = os.path.join(output_folder_base,relative_fn)
    try:
        os.makedirs(output_folder_video,exist_ok=True)
    except Exception:
        output_folder_clean = clean_path(output_folder_video)
        print('Warning: failed to create folder {}, trying {}'.format(
            output_folder_video,output_folder_clean))
        output_folder_video = output_folder_clean
        os.makedirs(output_folder_video,exist_ok=True)

    # Render frames
    # input_video_file = input_fn_absolute; output_folder = output_folder_video
    frame_filenames,fs = video_to_frames(input_fn_absolute,
                                         output_folder_video,
                                         overwrite=overwrite,
                                         every_n_frames=every_n_frames,
                                         verbose=verbose,
                                         quality=quality,
                                         max_width=max_width,
                                         frames_to_extract=frames_to_extract,
                                         allow_empty_videos=allow_empty_videos)

    return frame_filenames,fs


def video_folder_to_frames(input_folder,
                           output_folder_base,
                           recursive=True,
                           overwrite=True,
                           n_threads=1,
                           every_n_frames=None,
                           verbose=False,
                           parallelization_uses_threads=True,
                           quality=None,
                           max_width=None,
                           frames_to_extract=None,
                           allow_empty_videos=False,
                           relative_paths_to_process=None):
    """
    For every video file in input_folder, creates a folder within output_folder_base, and
    renders frame of that video to images in that folder.

    Args:
        input_folder (str): folder to process
        output_folder_base (str): root folder for output images; subfolders will be
            created for each input video
        recursive (bool, optional): whether to recursively process videos in [input_folder]
        overwrite (bool, optional): whether to overwrite existing frame images
        n_threads (int, optional): number of concurrent workers to use; set to <= 1 to disable
            parallelism
        every_n_frames (int or float, optional): sample every Nth frame starting from the first
            frame; if this is None or 1, every frame is extracted.  If this is a negative value,
            it's interpreted as a sampling rate in seconds, which is rounded to the nearest frame
            sampling rate.  Mutually exclusive with frames_to_extract.
        verbose (bool, optional): enable additional debug console output
        parallelization_uses_threads (bool, optional): whether to use threads (True) or
            processes (False) for parallelization; ignored if n_threads <= 1
        quality (int, optional): JPEG quality for frame output, from 0-100.  Defaults
            to the opencv default (typically 95).
        max_width (int, optional): resize frames to be no wider than [max_width]
        frames_to_extract (int, list of int, or dict, optional): extract this specific set of frames
            from each video; mutually exclusive with every_n_frames.  If all values are beyond the
            length of a video, no frames are extracted. Can also be a single int, specifying a single
            frame number.  In the special case where frames_to_extract is [], this function still
            reads video frame rates and verifies that videos are readable, but no frames are
            extracted. Can be a dict mapping relative paths to lists of frame numbers to extract different
            frames from each video.
        allow_empty_videos (bool, optional): just print a warning if a video appears to have no
            frames (by default, this is an error).
        relative_paths_to_process (list, optional): only process the relative paths on this
            list

    Returns:
        tuple: a length-3 tuple containing:
            - list of lists of frame filenames; the Nth list of frame filenames corresponds to
              the Nth video
            - list of video frame rates; the Nth value corresponds to the Nth video
            - list of video filenames
    """

    # Enumerate video files if necessary
    if relative_paths_to_process is None:
        if verbose:
            print('Enumerating videos in {}'.format(input_folder))
        input_files_full_paths = find_videos(input_folder,recursive=recursive)
        if verbose:
            print('Found {} videos in folder {}'.format(len(input_files_full_paths),input_folder))
        if len(input_files_full_paths) == 0:
            return [],[],[]

        input_files_relative_paths = [os.path.relpath(s,input_folder) for s in input_files_full_paths]
    else:
        input_files_relative_paths = relative_paths_to_process
        input_files_full_paths = [os.path.join(input_folder,fn) for fn in input_files_relative_paths]

    input_files_relative_paths = [s.replace('\\','/') for s in input_files_relative_paths]

    os.makedirs(output_folder_base,exist_ok=True)

    frame_filenames_by_video = []
    fs_by_video = []

    if n_threads == 1:

        # For each video
        #
        # input_fn_relative = input_files_relative_paths[0]
        for input_fn_relative in tqdm(input_files_relative_paths,desc='Video to frames'):

            # If frames_to_extract is a dict, get the specific frames for this video
            if isinstance(frames_to_extract, dict):
                frames_for_this_video = frames_to_extract.get(input_fn_relative, [])
            else:
                frames_for_this_video = frames_to_extract

            frame_filenames,fs = \
                _video_to_frames_for_folder(input_fn_relative,
                                            input_folder,
                                            output_folder_base,
                                            every_n_frames,
                                            overwrite,
                                            verbose,
                                            quality,
                                            max_width,
                                            frames_for_this_video,
                                            allow_empty_videos)
            frame_filenames_by_video.append(frame_filenames)
            fs_by_video.append(fs)

    else:

        pool = None
        results = None
        try:

            if parallelization_uses_threads:
                print('Starting a worker pool with {} threads'.format(n_threads))
                pool = ThreadPool(n_threads)
            else:
                print('Starting a worker pool with {} processes'.format(n_threads))
                pool = Pool(n_threads)

            if isinstance(frames_to_extract, dict):

                # For the dict case, we need to extract different frames from each video.

                # These arguments are the same for every iteration
                other_args = (input_folder, output_folder_base, every_n_frames, overwrite,
                              verbose, quality, max_width, allow_empty_videos)

                # The filename and list of frames to extract vary with each iteration
                args_for_pool = [(relative_fn, frames_to_extract.get(relative_fn, []), other_args)
                                 for relative_fn in input_files_relative_paths]

                results = list(tqdm(pool.imap(_video_to_frames_with_per_video_frames, args_for_pool),
                                    total=len(args_for_pool),desc='Video to frames'))

            else:

                process_video_with_options = partial(_video_to_frames_for_folder,
                                                     input_folder=input_folder,
                                                     output_folder_base=output_folder_base,
                                                     every_n_frames=every_n_frames,
                                                     overwrite=overwrite,
                                                     verbose=verbose,
                                                     quality=quality,
                                                     max_width=max_width,
                                                     frames_to_extract=frames_to_extract,
                                                     allow_empty_videos=allow_empty_videos)
                results = list(tqdm(pool.imap(process_video_with_options, input_files_relative_paths),
                                    total=len(input_files_relative_paths),desc='Video to frames'))

            # ...if we need to pass different frames for each video

        finally:

            if pool is not None:
                pool.close()
                pool.join()
                print('Pool closed and joined for video processing')

        # ...try/finally

        frame_filenames_by_video = [x[0] for x in results]
        fs_by_video = [x[1] for x in results]

    # ...if we're working on a single thread vs. multiple workers

    return frame_filenames_by_video,fs_by_video,input_files_full_paths

# ...def video_folder_to_frames(...)


class FrameToVideoOptions:
    """
    Options controlling the conversion of frame-level results to video-level results via
    frame_results_to_video_results()
    """

    def __init__(self):

        #: One-indexed indicator of which frame-level confidence value to use to determine detection confidence
        #: for the whole video, i.e. "1" means "use the confidence value from the highest-confidence frame"
        self.nth_highest_confidence = 1

        #: Should we include just a single representative frame result for each video (default), or
        #: every frame that was processed?
        self.include_all_processed_frames = False

        #: What to do if a file referred to in a .json results file appears not to be a
        #: video; can be 'error' or 'skip_with_warning'
        self.non_video_behavior = 'error'

        #: Are frame rates required?
        self.frame_rates_are_required = False

        #: Enable additional debug output
        self.verbose = False


def frame_results_to_video_results(input_file,
                                   output_file,
                                   options=None,
                                   video_filename_to_frame_rate=None):
    """
    Given an MD results file produced at the *frame* level, corresponding to a directory
    created with video_folder_to_frames, maps those frame-level results back to the
    video level for use in Timelapse.

    Preserves everything in the input .json file other than the images.

    Args:
        input_file (str): the frame-level MD results file to convert to video-level results
        output_file (str): the .json file to which we should write video-level results
        options (FrameToVideoOptions, optional): parameters for converting frame-level results
            to video-level results, see FrameToVideoOptions for details
        video_filename_to_frame_rate (dict, optional): maps (relative) video path names to frame
            rates, used only to populate the output file
    """

    if options is None:
        options = FrameToVideoOptions()

    if options.frame_rates_are_required:
        assert video_filename_to_frame_rate is not None, \
            'You specified that frame rates are required, but you did not ' + \
            'supply video_filename_to_frame_rate'

    # Load results
    with open(input_file,'r') as f:
        input_data = json.load(f)

    images = input_data['images']
    detection_categories = input_data['detection_categories']


    ## Break into videos

    video_to_frame_info = defaultdict(list)

    # im = images[0]
    for im in tqdm(images):

        fn = im['file']
        video_name = os.path.dirname(fn)

        if not is_video_file(video_name):

            if options.non_video_behavior == 'error':
                raise ValueError('{} is not a video file'.format(video_name))
            elif options.non_video_behavior == 'skip_with_warning':
                print('Warning: {} is not a video file'.format(video_name))
                continue
            else:
                raise ValueError('Unrecognized non-video handling behavior: {}'.format(
                    options.non_video_behavior))

        # Attach video-specific fields to the output, specifically attach the frame
        # number to both the video and each detection.
        frame_number = _filename_to_frame_number(fn)
        im['frame_number'] = frame_number
        for detection in im['detections']:
            detection['frame_number'] = frame_number

        video_to_frame_info[video_name].append(im)

    # ...for each frame referred to in the results file

    print('Found {} unique videos in {} frame-level results'.format(
        len(video_to_frame_info),len(images)))

    output_images = []


    ## For each video...

    # video_name = list(video_to_frame_info.keys())[0]
    for video_name in tqdm(video_to_frame_info):

        # Prepare the output representation for this video
        im_out = {}
        im_out['file'] = video_name

        if (video_filename_to_frame_rate is not None):

            if video_name not in video_filename_to_frame_rate:

                s = 'Could not determine frame rate for {}'.format(video_name)
                if options.frame_rates_are_required:
                    raise ValueError(s)
                elif options.verbose:
                    print('Warning: {}'.format(s))

            if video_name in video_filename_to_frame_rate:
                im_out['frame_rate'] = video_filename_to_frame_rate[video_name]

        # Find all detections for this video
        all_detections_this_video = []

        frames = video_to_frame_info[video_name]

        # frame = frames[0]
        for frame in frames:
            if ('detections' in frame) and (frame['detections'] is not None):
                all_detections_this_video.extend(frame['detections'])

        # Should we keep detections for all frames?
        if (options.include_all_processed_frames):

            im_out['detections'] = all_detections_this_video

        # ...or should we keep just a canonical detection for each category?
        else:

            canonical_detections = []

            # category_id = list(detection_categories.keys())[0]
            for category_id in detection_categories:

                category_detections = [det for det in all_detections_this_video if \
                                       det['category'] == category_id]

                # Find the nth-highest-confidence video to choose a confidence value
                if len(category_detections) >= options.nth_highest_confidence:

                    category_detections_by_confidence = sorted(category_detections,
                                                               key = lambda i: i['conf'],reverse=True)
                    canonical_detection = category_detections_by_confidence[options.nth_highest_confidence-1]
                    canonical_detections.append(canonical_detection)

            im_out['detections'] = canonical_detections

            # 'max_detection_conf' is no longer included in output files by default
            if False:
                im_out['max_detection_conf'] = 0
                if len(canonical_detections) > 0:
                    confidences = [d['conf'] for d in canonical_detections]
                    im_out['max_detection_conf'] = max(confidences)

        # ...if we're keeping output for all frames / canonical frames

        output_images.append(im_out)

    # ...for each video

    output_data = input_data
    output_data['images'] = output_images
    s = json.dumps(output_data,indent=1)

    # Write the output file
    with open(output_file,'w') as f:
        f.write(s)

# ...def frame_results_to_video_results(...)


#%% Test drivers

if False:

    pass

    #%% Constants

    input_folder = r'G:\temp\usu-long\data'
    frame_folder_base = r'g:\temp\usu-long-single-frames'
    assert os.path.isdir(input_folder)


    #%% Split videos into frames

    frame_filenames_by_video,fs_by_video,video_filenames = \
        video_folder_to_frames(input_folder,
                               frame_folder_base,
                               recursive=True,
                               overwrite=True,
                               n_threads=10,
                               every_n_frames=None,
                               verbose=True,
                               parallelization_uses_threads=True,
                               quality=None,
                               max_width=None,
                               frames_to_extract=150)


    #%% Constants for detection tests

    detected_frame_folder_base = r'e:\video_test\detected_frames'
    rendered_videos_folder_base = r'e:\video_test\rendered_videos'
    os.makedirs(detected_frame_folder_base,exist_ok=True)
    os.makedirs(rendered_videos_folder_base,exist_ok=True)
    results_file = r'results.json'
    confidence_threshold = 0.75


    #%% Load detector output

    with open(results_file,'r') as f:
        detection_results = json.load(f)
    detections = detection_results['images']
    detector_label_map = detection_results['detection_categories']
    for d in detections:
        d['file'] = d['file'].replace('\\','/').replace('video_frames/','')


    #%% List image files, break into folders

    frame_files = path_utils.find_images(frame_folder_base,True)
    frame_files = [s.replace('\\','/') for s in frame_files]
    print('Enumerated {} total frames'.format(len(frame_files)))

    # Find unique folders
    folders = set()
    # fn = frame_files[0]
    for fn in frame_files:
        folders.add(os.path.dirname(fn))
    folders = [s.replace('\\','/') for s in folders]
    print('Found {} folders for {} files'.format(len(folders),len(frame_files)))


    #%% Render detector frames

    # folder = list(folders)[0]
    for folder in folders:

        frame_files_this_folder = [fn for fn in frame_files if folder in fn]
        folder_relative = folder.replace((frame_folder_base + '/').replace('\\','/'),'')
        detection_results_this_folder = [d for d in detections if folder_relative in d['file']]
        print('Found {} detections in folder {}'.format(len(detection_results_this_folder),folder))
        assert len(frame_files_this_folder) == len(detection_results_this_folder)

        rendered_frame_output_folder = os.path.join(detected_frame_folder_base,folder_relative)
        os.makedirs(rendered_frame_output_folder,exist_ok=True)

        # d = detection_results_this_folder[0]
        for d in tqdm(detection_results_this_folder):

            input_file = os.path.join(frame_folder_base,d['file'])
            output_file = os.path.join(detected_frame_folder_base,d['file'])
            os.makedirs(os.path.dirname(output_file),exist_ok=True)
            vis_utils.draw_bounding_boxes_on_file(input_file,output_file,d['detections'],
                                                  confidence_threshold)

        # ...for each file in this folder

    # ...for each folder


    #%% Render output videos

    # folder = list(folders)[0]
    for folder in tqdm(folders):

        folder_relative = folder.replace((frame_folder_base + '/').replace('\\','/'),'')
        rendered_detector_output_folder = os.path.join(detected_frame_folder_base,folder_relative)
        assert os.path.isdir(rendered_detector_output_folder)

        frame_files_relative = os.listdir(rendered_detector_output_folder)
        frame_files_absolute = [os.path.join(rendered_detector_output_folder,s) \
                                for s in frame_files_relative]

        output_video_filename = os.path.join(rendered_videos_folder_base,folder_relative)
        os.makedirs(os.path.dirname(output_video_filename),exist_ok=True)

        original_video_filename = output_video_filename.replace(
            rendered_videos_folder_base,input_folder)
        assert os.path.isfile(original_video_filename)
        fs = get_video_fs(original_video_filename)

        frames_to_video(frame_files_absolute, fs, output_video_filename)

    # ...for each video



"""

map_lila_taxonomy_to_wi_taxonomy.py

Loads the LILA category mapping (in which taxonomy information comes from an
iNat taxonomy snapshot) and tries to map each class to the Wildlife Insights taxonomy.

What I do when I run this:

* Run all cells up to "Manual review of redundant taxa"

* Run the cell called "Go through each redundant taxon, one at a time" for each of the redundant taxa (~25 as of 2026.01), and update or create a corresponding entry in the "Decisions for redundant taxa" cell

* Run cells up to and including "Map LILA categories to WI categories", there will be errors, edit lila_to_wi_supplementary_mapping_file.csv to fix them

* Run through the last cell

* Upload lila_wi_mapping_table.csv to LILA

"""

#%% Constants and imports

import numpy as np
import json
import os

from tqdm import tqdm

from megadetector.data_management.lila.lila_common import \
    read_lila_taxonomy_mapping, read_wildlife_insights_taxonomy_mapping


#%% Prevent execution during infrastructural imports

if False:

    #%%

    lila_local_base = os.path.expanduser('~/lila')

    metadata_dir = os.path.join(lila_local_base, 'metadata')
    os.makedirs(metadata_dir, exist_ok=True)

    # Created by get_lila_category_list.py... contains counts for each category
    category_list_dir = os.path.join(lila_local_base, 'lila_categories_list')
    lila_dataset_to_categories_file = os.path.join(
        category_list_dir, 'lila_dataset_to_categories.json')

    # This is a manually-curated file used to store mappings that had to be made manually
    #
    # Specifically, this maps names that appear in LILA to WI names.  Some of these are scientific
    # names getting mapped from a taxonomy level that SpeciesNet doesn't use (e.g. mapping the pteromyini
    # tribe (flying squirrel) up to the sciuridae family (squirrels)), some are various versions of
    # "unknown" (e.g. "problem", "foggy weather").
    lila_to_wi_supplementary_mapping_file = os.path.expanduser(
        '~/git/MegaDetector/megadetector/taxonomy_mapping/lila_to_wi_supplementary_mapping_file.csv')

    assert os.path.isfile(lila_dataset_to_categories_file)

    # This is the main output file from this whole process
    wi_mapping_table_file = os.path.join(lila_local_base,'lila_wi_mapping_table.csv')

    id_column = 'uniqueIdentifier' # 'id'


    #%% Load category and taxonomy files

    with open(lila_dataset_to_categories_file, 'r') as f:
        lila_dataset_to_categories = json.load(f)

    lila_taxonomy_df = read_lila_taxonomy_mapping(metadata_dir)

    wi_taxonomy_df = read_wildlife_insights_taxonomy_mapping(metadata_dir)


    #%% Pull everything out of pandas

    lila_taxonomy = lila_taxonomy_df.to_dict('records')
    wi_taxonomy = wi_taxonomy_df.to_dict('records')


    #%% Cache WI taxonomy lookups

    def _is_empty_wi_item(v):
        if isinstance(v, str):
            return len(v) == 0
        elif v is None:
            return True
        else:
            assert isinstance(v, float) and np.isnan(v), 'Invalid item: {}'.format(str(v))
            return True

    def _taxonomy_items_equal(a, b):
        if isinstance(a, str) and (not isinstance(b, str)):
            return False
        if isinstance(b, str) and (not isinstance(a, str)):
            return False
        if (not isinstance(a, str)) or (not isinstance(b, str)):
            assert (a is None and b is None) or (isinstance(a, float) and isinstance(b, float))
            return True
        return a == b

    for taxon in wi_taxonomy:
        taxon['taxon_name'] = None

    from collections import defaultdict
    wi_taxon_name_to_taxa = defaultdict(list)

    # This is just a handy lookup table that we'll use to debug mismatches
    wi_common_name_to_taxon = {}

    blank_taxon_name = 'blank'
    blank_taxon = None

    animal_taxon_name = 'animal'
    animal_taxon = None

    unknown_taxon_name = 'unknown'
    unknown_taxon = None

    ignore_taxa = set(['No CV Result', 'CV Needed', 'CV Failed'])

    known_problematic_taxon_ids = ['f94e6d97-59cf-4d38-a05a-a75efdd2863b']

    human_taxa = []

    # taxon = wi_taxonomy[21653]; print(taxon)
    for taxon in tqdm(wi_taxonomy):

        taxon_name = None

        assert taxon['taxonomyType'] == 'object' or taxon['taxonomyType'] == 'biological'

        for k in taxon.keys():
            v = taxon[k]
            if isinstance(v,str):
                taxon[k] = v.strip()

        if taxon['commonNameEnglish'] in ignore_taxa:
            continue

        if isinstance(taxon['commonNameEnglish'], str):

            wi_common_name_to_taxon[taxon['commonNameEnglish'].strip(
            ).lower()] = taxon

            special_taxon = False

            # Look for keywords that don't refer to specific taxa: blank/animal/unknown
            if taxon['commonNameEnglish'].strip().lower() == blank_taxon_name:
                blank_taxon = taxon
                special_taxon = True

            elif taxon['commonNameEnglish'].strip().lower() == animal_taxon_name:
                animal_taxon = taxon
                special_taxon = True

            elif taxon['commonNameEnglish'].strip().lower() == unknown_taxon_name:
                unknown_taxon = taxon
                special_taxon = True

            if special_taxon:
                taxon_name = taxon['commonNameEnglish'].strip().lower()
                taxon['taxon_name'] = taxon_name
                wi_taxon_name_to_taxa[taxon_name].append(taxon)
                continue

        # Do we have a species name?
        if not _is_empty_wi_item(taxon['species']):

            # If 'species' is populated, 'genus' should always be populated
            assert not _is_empty_wi_item(taxon['genus'])

            taxon_name = (taxon['genus'].strip() + ' ' +
                        taxon['species'].strip()).strip().lower()
            assert not _is_empty_wi_item(taxon['class']) and \
                not _is_empty_wi_item(taxon['order']) and \
                not _is_empty_wi_item(taxon['family'])

        elif not _is_empty_wi_item(taxon['genus']):

            assert not _is_empty_wi_item(taxon['class']) and \
                not _is_empty_wi_item(taxon['order']) and \
                not _is_empty_wi_item(taxon['family'])
            taxon_name = taxon['genus'].strip().lower()

        elif not _is_empty_wi_item(taxon['family']):

            assert not _is_empty_wi_item(taxon['class']) and \
                not _is_empty_wi_item(taxon['order'])
            taxon_name = taxon['family'].strip().lower()

        elif not _is_empty_wi_item(taxon['order']):

            assert not _is_empty_wi_item(taxon['class'])
            taxon_name = taxon['order'].strip().lower()

        elif not _is_empty_wi_item(taxon['class']):

            taxon_name = taxon['class'].strip().lower()

        if taxon_name is not None:
            assert taxon['taxonomyType'] == 'biological'
        else:
            assert taxon['taxonomyType'] == 'object'
            taxon_name = taxon['commonNameEnglish'].strip().lower()

        if taxon_name in wi_taxon_name_to_taxa:
            if taxon[id_column] in known_problematic_taxon_ids:
                print('Skipping known-problematic taxon ID {}: {}'.format(
                    taxon[id_column],str(taxon)))
            else:
                previous_taxa = wi_taxon_name_to_taxa[taxon_name]
                for previous_taxon in previous_taxa:
                    for level in ['class', 'order', 'family', 'genus', 'species']:
                        error_string = 'Error: taxon {} appeared previously in {} {} (as {}), now in {} {}'.format(
                            taxon_name,
                            level,previous_taxon[level],
                            previous_taxon['taxon_name'],
                            level,taxon[level])
                        assert _taxonomy_items_equal(previous_taxon[level], taxon[level]), error_string

        taxon['taxon_name'] = taxon_name
        if taxon_name == 'homo sapiens':
            human_taxa.append(taxon)
        wi_taxon_name_to_taxa[taxon_name].append(taxon)

    # ...for each taxon

    assert unknown_taxon is not None
    assert animal_taxon is not None
    assert blank_taxon is not None


    #%% Find redundant taxa

    taxon_names_with_multiple_entries = []
    for wi_taxon_name in wi_taxon_name_to_taxa:
        if len(wi_taxon_name_to_taxa[wi_taxon_name]) > 1:
            taxon_names_with_multiple_entries.append(wi_taxon_name)

    print('{} names have multiple entries:\n'.format(len(taxon_names_with_multiple_entries)))

    # wi_taxon_name = taxon_names_with_multiple_entries[0]
    for wi_taxon_name in taxon_names_with_multiple_entries:

        # Look at disagreements in the different entries for this taxon name
        taxa = wi_taxon_name_to_taxa[wi_taxon_name]
        keys = set(taxa[0].keys())

        for taxon in taxa:
            assert set(taxon.keys()) == keys

        disagreements = defaultdict(list)

        for k in keys:
            if k.lower() in ('createdat','updatedat','uniqueidentifier','id','referenceUrl','authority'):
                continue
            values = set()
            for taxon in taxa:
                v = taxon[k]
                if isinstance(v,float) and np.isnan(v):
                    v = 'nan'
                values.add(v)
            # Is there more than one distinct value for this field?
            if len(values) > 1:
                disagreements[k] = values
        # ...for each field name

        print('  {}'.format(wi_taxon_name))

        for k in disagreements:
            print('    {}: {}'.format(k,str(disagreements[k])))


    #%% Manual review of redundant taxa

    if False:

        pass

        #%% Go through each redundant taxon, one at a time

        import clipboard


        i_taxon = 12

        if False:
            s = 'sericornis'
            for i_taxon,taxon_name in enumerate(taxon_names_with_multiple_entries):
                if s == taxon_name:
                    break

        s = taxon_names_with_multiple_entries[i_taxon]
        taxa = wi_taxon_name_to_taxa[s]
        for i_option,t in enumerate(taxa):
            print('Index {}:'.format(i_option))
            for k in t.keys():
                print('{}: {}'.format(k,t[k]))
            print('')
            # print(t,end='\n\n')

        i_selection = 0
        if i_selection >= len(taxa):
            i_selection = 0

        """
        # "helmeted guineafowl" vs "domestic guineafowl"
        taxon_name_to_preferred_taxon_id['numida meleagris'] = '83133617-8358-4910-82ee-4c23e40ba3dc' # 2005826
        """

        common_names = [t['commonNameEnglish'] for t in taxa]

        # Replace NaN entries in "common_names" with the string "nan"
        for i_cn in range(len(common_names)):
            if not isinstance(common_names[i_cn], str):
                common_names[i_cn] = 'nan'

        selected_common_name = common_names[i_selection]
        non_selected_names = [s for s in common_names if s != selected_common_name]
        non_selected_names = ', '.join(non_selected_names)
        output_string = '# "{}" (instead of {})'.format(selected_common_name,
                                                        non_selected_names)
        selected_entry = taxa[i_selection]
        # taxon_name_to_preferred_taxon_id['numida meleagris'] = '83133617-8358-4910-82ee-4c23e40ba3dc' # 2005826
        output_string += \
            "\ntaxon_name_to_preferred_taxon_id['{}'] = '{}' # {}".format(
                selected_entry['taxon_name'],
                selected_entry['uniqueIdentifier'],
                selected_entry['id'])

        print(output_string)
        clipboard.copy(output_string)

        #%%

    # ...if False (manual review)


    #%% Decisions for redundant taxa

    taxon_name_to_preferred_taxon_id = {}

    # "Helmeted Guineafowl" (instead of Domestic Guineafowl)
    taxon_name_to_preferred_taxon_id['numida meleagris'] = '83133617-8358-4910-82ee-4c23e40ba3dc' # 2005826

    # "Wild Turkey" (instead of Domestic Turkey)
    taxon_name_to_preferred_taxon_id['meleagris gallopavo'] = '94d8284e-6193-43b5-b80c-9029cf951675' # 2006047

    # "Human" (instead of Horseback Rider, Human-Faces, Human-Herder, Homo Species, Human-Camera Trapper, Human-Hunter,
    # Human-Researcher, Human-Tourist, Human-Park Ranger, Human-horseback rider, Human-Resident, Human-Maintenance crew
    # Human-Pedestrian, Human - Soldier, Human - Biker, Human-Skier, Human-Snowshoer, Human-Hiker)
    taxon_name_to_preferred_taxon_id['homo sapiens'] = '990ae9dd-7a59-4344-afcb-1b7b21368000' # 2002045

    # "Domestic Dog" (instead of Dog-on-leash)
    taxon_name_to_preferred_taxon_id['canis familiaris'] = '3d80f1d6-b1df-4966-9ff4-94053c7a902a' # 2021548

    # "Mammal" (instead of Small Mammal)
    taxon_name_to_preferred_taxon_id['mammalia'] = 'f2d233e3-80e3-433d-9687-e29ecc7a467a' # 2021108

    # "gastropods" (instead of Snail species, Slug species)
    taxon_name_to_preferred_taxon_id['gastropoda'] = '5b91403b-7361-4bb0-bf42-9a1687a88198' # 2021755

    # "Hispaniolan Mango" (instead of nan)
    taxon_name_to_preferred_taxon_id['anthracothorax dominicus'] = 'f94e6d97-59cf-4d38-a05a-a75efdd2863b' # 3

    # "Lizards and Snakes" (instead of Serpentes Suborder, Lacertilia Suborder)
    taxon_name_to_preferred_taxon_id['squamata'] = '0af344ad-6657-42c4-85d8-03fd2106a22a' # 2021551

    # "Beautiful Firetail" (instead of )
    taxon_name_to_preferred_taxon_id['stagonopleura bella'] = '7fec8e7e-fd3b-4d7f-99fd-3ade6f3bbaa5' # 2021939

    # "Rodent" (instead of Mouse species, Woodrat or Rat or Mouse Species, Woodrat or Rat Species)
    taxon_name_to_preferred_taxon_id['rodentia'] = '90d950db-2106-4bd9-a4c1-777604c3eada' # 2021173

    # "Yellow Crowned-wagtail" (instead of Yellow Wagtail)
    taxon_name_to_preferred_taxon_id['motacilla flava'] = 'a51fd282-2db6-4945-bebe-d64e52153bcf' # 2016200

    # "Dremomys Species" (instead of Dremomys Genus)
    taxon_name_to_preferred_taxon_id['dremomys'] = '1507d153-af11-46f1-bfb8-77918d035ab3' # 2019370

    # "Snowmobile" (instead of )
    taxon_name_to_preferred_taxon_id['snowmobile'] = 'c091eeb7-d1c5-4779-988f-82dd6fa12aae' # 2021826

    # "Elk" (instead of Domestic Elk)
    taxon_name_to_preferred_taxon_id['cervus canadensis'] = 'c5ce946f-8f0d-4379-992b-cc0982381f5e' # 2021592

    # "American Bison" (instead of Domestic Bison)
    taxon_name_to_preferred_taxon_id['bison bison'] = '539ebd55-081b-429a-9ae6-5a6a0f6999d4' # 2021593

    # "Cricetidae Family" (instead of Arvicolinae Subfamily)
    taxon_name_to_preferred_taxon_id['cricetidae'] = '523439f4-dee3-4f41-afea-edc0c891ef9c' # 2021115

    # "Heteromyidae Family" (instead of Perognathinae Subfamily)
    taxon_name_to_preferred_taxon_id['heteromyidae'] = 'a181f9da-8101-45dc-8bf4-06d6dcbebaef' # 2021148

    # "Weasel Family" (instead of Mustelinae Subfamily)
    taxon_name_to_preferred_taxon_id['mustelidae'] = '80d82668-81f1-4ffe-96a5-2325ce78ecf1' # 2021337

    # "Anniella Species" (instead of Aniella species)
    taxon_name_to_preferred_taxon_id['anniella'] = '46fd4962-c310-4a4a-833e-fa5fd068caa9' # 2021645

    # "Southern Sand Frog" (instead of )
    taxon_name_to_preferred_taxon_id['tomopterna adiastola'] = 'd8e212e7-6bd5-4539-a7f8-4eb43eba97d1' # 2021829

    # "Aspidoscelis Species" (instead of Aspidocelis species)
    taxon_name_to_preferred_taxon_id['aspidoscelis'] = 'd5bbf997-d2b4-4d4f-bcca-4b0b2e867e97' # 2021651

    # "Phrynosomatidae Family" (instead of Sceloporus/Uta species)
    taxon_name_to_preferred_taxon_id['phrynosomatidae'] = 'f38cb662-8569-48cd-b93b-34d20310b989' # 2021691

    # "White-rumped Shama" (instead of )
    taxon_name_to_preferred_taxon_id['copsychus malabaricus'] = '15ecc823-cbf1-4ada-bc5d-fc5a69cf1574' # 2015152

    # "Rosy Boa" (instead of Northern Three-lined Boa)
    taxon_name_to_preferred_taxon_id['lichanura orcutti'] = '7c373a44-7add-430d-88b0-197b7673459a' # 1107

    # "Comb-eared Skinks" (instead of Ctenotus genus)
    taxon_name_to_preferred_taxon_id['ctenotus'] = 'eae2e6df-e515-4b27-963e-3bb267205760' # 1180

    # "Agamidae Family" (instead of Dragon)
    taxon_name_to_preferred_taxon_id['agamidae'] = 'b3a3aaf6-dd72-4fdc-95ca-bcb282bca698' # 1271

    # "Sericornis Species" (instead of Scrubwren Species)
    taxon_name_to_preferred_taxon_id['sericornis'] = 'ad82c0ac-df48-4028-bf71-d2b2f4bc4129' # 2020133

    print('Defined preferences for {} taxa'.format(
        len(taxon_name_to_preferred_taxon_id)))

    assert len(taxon_name_to_preferred_taxon_id) <= \
        len(taxon_names_with_multiple_entries)

    for s in taxon_names_with_multiple_entries:
        if s not in taxon_name_to_preferred_taxon_id:
            print('Warning: no mapping for {}'.format(s))


    #%% Reduce each set of redundant taxa to a single choice

    # taxon_name = list(taxon_name_to_preferred_taxon_id.keys())[0]
    for taxon_name in taxon_name_to_preferred_taxon_id.keys():

        candidate_taxa = wi_taxon_name_to_taxa[taxon_name]

        # If we've gotten this far, we should be choosing from multiple taxa.
        #
        # This will become untrue if any of these are resolved later, at which point we should
        # remove them from taxon_name_to_preferred_id
        assert len(candidate_taxa) > 1, 'Only one taxon available for {}'.format(taxon_name)

        # Choose the preferred taxa
        selected_taxa = [t for t in candidate_taxa if t[id_column] == \
                        taxon_name_to_preferred_taxon_id[taxon_name]]
        assert len(selected_taxa) == 1
        wi_taxon_name_to_taxa[taxon_name] = selected_taxa

    wi_taxon_name_to_taxon = {}

    for taxon_name in wi_taxon_name_to_taxa.keys():
        taxa = wi_taxon_name_to_taxa[taxon_name]
        assert len(taxa) == 1
        wi_taxon_name_to_taxon[taxon_name] = taxa[0]


    #%% Read supplementary mappings

    with open(lila_to_wi_supplementary_mapping_file, 'r') as f:
        lines = f.readlines()

    # Discard header
    lines = lines[1:]
    lines = [s.strip() for s in lines if len(s.strip()) > 0]

    supplementary_lila_query_to_wi_query = {}

    for line in lines:
        assert '\t' not in line
        # Each line is [lila query],[WI taxon name],[notes]
        tokens = line.strip().split(',')
        assert len(tokens) == 3
        lila_query = tokens[0].strip().lower()
        wi_taxon_name = tokens[1].strip().lower()
        if wi_taxon_name not in wi_taxon_name_to_taxa:
            raise ValueError('{} not in WI taxonomy'.format(wi_taxon_name))
        supplementary_lila_query_to_wi_query[lila_query] = wi_taxon_name


    #%% Map LILA categories to WI categories

    mismatches = set()
    mismatches_with_common_mappings = set()
    supplementary_mappings = set()

    all_searches = set()

    # Must be ordered from kingdom --> species
    lila_taxonomy_levels = ['kingdom', 'phylum', 'subphylum', 'superclass', 'class', 'subclass',
                            'infraclass', 'superorder', 'order', 'suborder', 'infraorder',
                            'superfamily', 'family', 'subfamily', 'tribe', 'genus', 'species']

    unknown_queries = set(
        ['unidentifiable', 'other', 'unidentified', 'unknown', 'unclassifiable', 'no cv result'])
    blank_queries = set(['empty'])
    # This makes sense in the context of the WI taxonomy, which doesn't include invertebrates
    animal_queries = set(['animalia','vertebrata'])

    print_common_name_matches = True
    print_supplementary_matches = False

    lila_dataset_category_to_wi_taxon = {}

    # i_taxon = 0; lila_taxon = lila_taxonomy[i_taxon]; print(taxon)
    for i_taxon, lila_taxon in enumerate(lila_taxonomy):

        # What is the taxon name we should search for in the WI taxonomy?
        query = None

        lila_dataset_category = lila_taxon['dataset_name'] + ':' + lila_taxon['query']

        # Go from kingdom --> species, choosing the lowest-level description as the query
        for level in lila_taxonomy_levels:
            if isinstance(lila_taxon[level], str):
                query = lila_taxon[level]
                all_searches.add(query)

        # If we didn't find a taxon name, that means this is a non-taxonomic entity, like "car"
        if query is None:
            query = lila_taxon['query']

        wi_taxon = None

        if query in unknown_queries:

            wi_taxon = unknown_taxon

        elif query in blank_queries:

            wi_taxon = blank_taxon

        elif query in animal_queries:

            wi_taxon = animal_taxon

        elif query in wi_taxon_name_to_taxon:

            wi_taxon = wi_taxon_name_to_taxon[query]

        elif query in supplementary_lila_query_to_wi_query:

            wi_taxon = wi_taxon_name_to_taxon[supplementary_lila_query_to_wi_query[query]]
            supplementary_mappings.add(query)
            if print_supplementary_matches:
                print('Made a supplementary mapping from {} to {}'.format(query,wi_taxon['taxon_name']))

        else:

            # print('No match for {}'.format(query))
            lila_common_name = lila_taxon['common_name']

            if lila_common_name in wi_common_name_to_taxon:
                wi_taxon = wi_common_name_to_taxon[lila_common_name]
                wi_common_name = wi_taxon['commonNameEnglish']
                wi_taxon_name = wi_taxon['taxon_name']
                if print_common_name_matches:
                    print('LILA common name {} maps to WI taxon {} ({})'.format(lila_common_name,
                                                                                wi_taxon_name,
                                                                                wi_common_name))
                mismatches_with_common_mappings.add(query)

            else:

                mismatches.add(query)

        lila_dataset_category_to_wi_taxon[lila_dataset_category] = wi_taxon

    # ...for each LILA taxon

    print('Of {} entities, there are {} mismatches ({} mapped by common name) ({} mapped by supplementary mapping file)'.format(
        len(all_searches), len(mismatches), len(mismatches_with_common_mappings), len(supplementary_mappings)))

    assert len(mismatches) == 0


    #%% Write mapping file

    if not os.path.isfile(lila_to_wi_supplementary_mapping_file):
        print('Creating mapping file {}'.format(
            lila_to_wi_supplementary_mapping_file))
        with open(lila_to_wi_supplementary_mapping_file, 'w') as f:
            for query in mismatches:
                f.write(query + ',' + '\n')
    else:
        print('{} exists, not re-writing'.format(lila_to_wi_supplementary_mapping_file))


    #%% Build a dictionary from LILA dataset names and categories to LILA taxa

    lila_dataset_category_to_lila_taxon = {}

    # i_d = 0; d = lila_taxonomy[i_d]
    for i_d,d in enumerate(lila_taxonomy):
        lila_dataset_category = d['dataset_name'] + ':' + d['query']
        assert lila_dataset_category not in lila_dataset_category_to_lila_taxon
        lila_dataset_category_to_lila_taxon[lila_dataset_category] = d


    #%% Write mapping table file

    with open(wi_mapping_table_file,'w') as f:

        f.write('lila_dataset_name,lila_category_name,wi_guid,wi_taxon_name,wi_common,count\n')

        # dataset_name = list(lila_dataset_to_categories.keys())[0]
        for dataset_name in lila_dataset_to_categories.keys():

            if '_bbox' in dataset_name:
                continue

            dataset_categories = lila_dataset_to_categories[dataset_name]

            # dataset_category = dataset_categories[0]
            for category in dataset_categories:

                lila_dataset_category = dataset_name + ':' + category['name'].strip().lower()
                if '#' in lila_dataset_category:
                    continue
                assert lila_dataset_category in lila_dataset_category_to_lila_taxon
                assert lila_dataset_category in lila_dataset_category_to_wi_taxon
                assert 'count' in category

                wi_taxon = lila_dataset_category_to_wi_taxon[lila_dataset_category]

                # Write out the dataset name, category name, WI GUID, WI scientific name, WI common name,
                # and count
                s = f"{dataset_name},{category['name']},{wi_taxon['uniqueIdentifier']},"+\
                    f"{wi_taxon['taxon_name']},{wi_taxon['commonNameEnglish']},{category['count']}\n"
                f.write(s)

            # ...for each category in this dataset

        # ...for each dataset

    # ...with open()

    #%%

# ...if False (prevent execution at import)



"""

preview_lila_taxonomy.py

Does some consistency-checking on the LILA taxonomy file, and generates
an HTML preview page that we can use to determine whether the mappings
make sense.

"""

#%% Imports and constants

from tqdm import tqdm

import os
import pandas as pd

# lila_taxonomy_file = r"c:\git\agentmorrisprivate\lila-taxonomy\lila-taxonomy-mapping.csv"
lila_taxonomy_file = os.path.expanduser('~/lila/lila_additions_2026.03.18.csv')

preview_base = os.path.expanduser('~/lila/lila_taxonomy_preview')
os.makedirs(preview_base,exist_ok=True)
html_output_file = os.path.join(preview_base,'index.html')


#%% Support functions

def parse_taxonomy_string(taxonomy_string):

    taxonomic_match = eval(taxonomy_string)
    matched_entity = taxonomic_match[0]
    assert len(matched_entity) == 4

    level = matched_entity[1]

    scientific_name = matched_entity[2]

    common_names = matched_entity[3]
    if len(common_names) == 1:
        common_name = common_names[0]
    else:
        common_name = str(common_names)

    return scientific_name,common_name,level,taxonomic_match

def taxonomy_string_to_common_name(taxonomy_string):
    _,cn,_,_ = parse_taxonomy_string(taxonomy_string)
    return cn

def taxonomy_string_to_scientific(taxonomy_string):
    sn,_,_,_ = parse_taxonomy_string(taxonomy_string)
    return sn

def taxonomy_string_to_level(taxonomy_string):
    _,_,level,_ = parse_taxonomy_string(taxonomy_string)
    return level


#%% Prepare taxonomy lookup

from megadetector.taxonomy_mapping.species_lookup import \
    initialize_taxonomy_lookup, get_preferred_taxonomic_match

initialize_taxonomy_lookup()


#%% Check for mappings that disagree with the taxonomy string

# For example, cases where the "level" column says "species", but the taxonomy string says it's a genus.

df = pd.read_csv(lila_taxonomy_file)

n_taxonomy_changes = 0

# Look for internal inconsistency
for i_row,row in df.iterrows():

    sn = row['scientific_name']
    if not isinstance(sn,str):
        continue

    ts = row['taxonomy_string']
    assert sn == taxonomy_string_to_scientific(ts)

    assert row['taxonomy_level'] == taxonomy_string_to_level(ts)

# Look for outdated mappings
taxonomy_preference = 'inat'

# i_row = 0; row = df.iloc[i_row]
for i_row,row in tqdm(df.iterrows(),total=len(df)):

    try:

        sn = row['scientific_name']
        if not isinstance(sn,str):
            continue

        m = get_preferred_taxonomic_match(sn,taxonomy_preference)
        assert m.scientific_name == sn

        ts = row['taxonomy_string']
        assert m.taxonomy_string[0:50] == ts[0:50], 'Mismatch for {}:\n\n{}\n\n{}\n'.format(
            row['dataset_name'],ts,m.taxonomy_string)

        if ts != m.taxonomy_string:
            n_taxonomy_changes += 1
            df.loc[i_row,'taxonomy_string'] = m.taxonomy_string

    except Exception as e:

        print('Error at row {}: {}'.format(i_row,str(e)))
        raise

# ...for each row

print('\nMade {} taxonomy changes'.format(n_taxonomy_changes))

# Optionally re-write
if False:
    df.to_csv(lila_taxonomy_file,header=True,index=False)


#%% List null mappings

# These should all be things like "empty", "unidentified", "fire", "car", etc.

# i_row = 0; row = df.iloc[i_row]
for i_row,row in df.iterrows():
    if (not isinstance(row['taxonomy_string'],str)) or (len(row['taxonomy_string']) == 0):
        print('No mapping for {}:{}'.format(row['dataset_name'],row['query']))


#%% List mappings with scientific names but no common names

for i_row,row in df.iterrows():
    cn = row['common_name']
    sn = row['scientific_name']
    ts = row['taxonomy_string']
    if (isinstance(ts,str)) and (len(ts) >= 0):
        if (not isinstance(cn,str)) or (len(cn) == 0):
            print('No mapping for {}:{}:{}'.format(row['dataset_name'],row['query'],row['scientific_name']))


#%% List mappings that map to different things in different data sets

import numpy as np
def isnan(x):
    if not isinstance(x,float):
        return False
    return np.isnan(x)

from collections import defaultdict
query_to_rows = defaultdict(list)

queries_with_multiple_mappings = set()

n_suppressed = 0

suppress_multiple_matches = [
    ['porcupine','Snapshot Camdeboo','Idaho Camera Traps'],
    ['porcupine','Snapshot Enonkishu','Idaho Camera Traps'],
    ['porcupine','Snapshot Karoo','Idaho Camera Traps'],
    ['porcupine','Snapshot Kgalagadi','Idaho Camera Traps'],
    ['porcupine','Snapshot Kruger','Idaho Camera Traps'],
    ['porcupine','Snapshot Mountain Zebra','Idaho Camera Traps'],
    ['porcupine','Snapshot Serengeti','Idaho Camera Traps'],

    ['porcupine','Snapshot Serengeti','Snapshot Mountain Zebra'],
    ['porcupine','Snapshot Serengeti','Snapshot Kruger'],
    ['porcupine','Snapshot Serengeti','Snapshot Kgalagadi'],
    ['porcupine','Snapshot Serengeti','Snapshot Karoo'],
    ['porcupine','Snapshot Serengeti','Snapshot Camdeboo'],

    ['porcupine','Snapshot Enonkishu','Snapshot Camdeboo'],
    ['porcupine','Snapshot Enonkishu','Snapshot Mountain Zebra'],
    ['porcupine','Snapshot Enonkishu','Snapshot Kruger'],
    ['porcupine','Snapshot Enonkishu','Snapshot Kgalagadi'],
    ['porcupine','Snapshot Enonkishu','Snapshot Karoo'],

    ['kudu','Snapshot Serengeti','Snapshot Mountain Zebra'],
    ['kudu','Snapshot Serengeti','Snapshot Kruger'],
    ['kudu','Snapshot Serengeti','Snapshot Kgalagadi'],
    ['kudu','Snapshot Serengeti','Snapshot Karoo'],
    ['kudu','Snapshot Serengeti','Snapshot Camdeboo'],

    ['fox','Caltech Camera Traps','Channel Islands Camera Traps'],
    ['fox','Idaho Camera Traps','Channel Islands Camera Traps'],
    ['fox','Idaho Camera Traps','Caltech Camera Traps'],

    ['pangolin','Snapshot Serengeti','SWG Camera Traps'],

    ['deer', 'Wellington Camera Traps', 'Idaho Camera Traps'],
    ['deer', 'Wellington Camera Traps', 'Caltech Camera Traps'],

    ['unknown cervid', 'WCS Camera Traps', 'Idaho Camera Traps']

]

for i_row,row in df.iterrows():

    query = row['query']
    taxonomy_string = row['taxonomy_string']

    for previous_i_row in query_to_rows[query]:

        previous_row = df.iloc[previous_i_row]
        assert previous_row['query'] == query
        query_match = False
        if isnan(row['taxonomy_string']):
            query_match = isnan(previous_row['taxonomy_string'])
        elif isnan(previous_row['taxonomy_string']):
            query_match = isnan(row['taxonomy_string'])
        else:
            query_match = previous_row['taxonomy_string'][0:10] == taxonomy_string[0:10]

        if not query_match:

            suppress = False

            # x = suppress_multiple_matches[-1]
            for x in suppress_multiple_matches:
                if x[0] == query and \
                    ( \
                    (x[1] == row['dataset_name'] and x[2] == previous_row['dataset_name']) \
                    or \
                    (x[2] == row['dataset_name'] and x[1] == previous_row['dataset_name']) \
                    ):
                    suppress = True
                    n_suppressed += 1
                    break

            if not suppress:
                print('Query {} in {} and {}:\n\n{}\n\n{}\n'.format(
                    query, row['dataset_name'], previous_row['dataset_name'],
                    taxonomy_string, previous_row['taxonomy_string']))

            queries_with_multiple_mappings.add(query)

    # ...for each row where we saw this query

    query_to_rows[query].append(i_row)

# ...for each row

print('Found {} queries with multiple mappings ({} occurrences suppressed)'.format(
    len(queries_with_multiple_mappings),n_suppressed))


#%% Verify that nothing "unidentified" maps to a species or subspecies

# E.g., "unidentified skunk" should never map to a specific species of skunk

allowable_unknown_species = [
    'unknown_tayra' # AFAIK this is a unique species, I'm not sure what's implied here
]

unk_queries = ['skunk']
for i_row,row in df.iterrows():

    query = row['query']
    level = row['taxonomy_level']

    if not isinstance(level,str):
        assert not isinstance(row['taxonomy_string'],str)
        continue

    if ( \
        'unidentified' in query or \
        ('unk' in query and ('skunk' not in query and 'chipmunk' not in query))\
        ) \
        and \
        ('species' in level):

        if query not in allowable_unknown_species:

            print('Warning: query {}:{} maps to {} {}'.format(
                row['dataset_name'],
                row['query'],
                row['taxonomy_level'],
                row['scientific_name']
                ))


#%% Make sure there are valid source and level values for everything with a mapping

for i_row,row in df.iterrows():
    if isinstance(row['scientific_name'],str):
        if 'source' in row:
            assert isinstance(row['source'],str)
        assert isinstance(row['taxonomy_level'],str)


#%% Find WCS mappings that aren't species or aren't the same as the input

# WCS used scientific names, so these remappings are slightly more controversial
# then the standard remappings.

# row = df.iloc[-500]
for i_row,row in df.iterrows():

    if not isinstance(row['scientific_name'],str):
        continue
    if 'WCS' not in row['dataset_name']:
        continue

    query = row['query']
    scientific_name = row['scientific_name']
    common_name = row['common_name']
    level = row['taxonomy_level']
    taxonomy_string = row['taxonomy_string']

    common_name_from_taxonomy = taxonomy_string_to_common_name(taxonomy_string)
    query_string = query.replace(' sp','')
    query_string = query_string.replace('unknown ','')

    # Anything marked "species" or "unknown" by definition doesn't map to a species,
    # so ignore these.
    if (' sp' not in query) and ('unknown' not in query) and \
        (level != 'species') and (level != 'subspecies'):
        print('WCS query {} ({}) remapped to {} {} ({})'.format(
            query,common_name,level,scientific_name,common_name_from_taxonomy))

    if query_string != scientific_name:
        pass
        # print('WCS query {} ({}) remapped to {} ({})'.format(
        #     query,common_name,scientific_name,common_names_from_taxonomy))


#%% Download sample images for all scientific names

# You might have to do this:
#
# pip install python-magic
# pip install python-magic-bin

# Takes ~1 minute per 10 rows

import os
from concurrent.futures import ThreadPoolExecutor, as_completed
from megadetector.taxonomy_mapping import simple_image_download


remapped_queries = {'papio':'papio+baboon',
                    'damaliscus lunatus jimela':'damaliscus lunatus',
                    'mazama':'genus+mazama',
                    'mirafra':'genus+mirafra'}

scientific_name_to_paths = {}
image_base = os.path.join(preview_base,'images')
images_per_query = 15
min_valid_images_per_query = 3
min_valid_image_size = 3000
n_download_workers = 8

# Build the list of queries that actually need downloading
queries_to_download = []
seen_queries = set()

for i_row,row in df.iterrows():

    s = row['scientific_name']

    if (not isinstance(s,str)) or (len(s)==0):
        continue

    query = s.replace(' ','+')

    if query in remapped_queries:
        query = remapped_queries[query]

    if query in seen_queries:
        continue
    seen_queries.add(query)

    query_folder = os.path.join(image_base,query)
    os.makedirs(query_folder,exist_ok=True)

    # Check whether we already have enough images for this query
    image_files = os.listdir(query_folder)
    image_fullpaths = [os.path.join(query_folder,fn) for fn in image_files]
    sizes = [os.path.getsize(p) for p in image_fullpaths]
    sizes_above_threshold = [x for x in sizes if x > min_valid_image_size]
    if len(sizes_above_threshold) > min_valid_images_per_query:
        print('Skipping query {}, already have {} images'.format(query,len(sizes_above_threshold)))
        continue

    queries_to_download.append(query)

# ...for each row in the mapping table

print('Need to download images for {} queries'.format(len(queries_to_download)))


def _download_images_for_query(query):
    """
    Download images for a single query using a per-call Downloader instance.
    """

    downloader = simple_image_download.Downloader()
    downloader.directory = image_base
    query_for_download = query.replace(' ','+')
    paths = downloader.download(query_for_download,
                                limit=images_per_query,
                                verbose=False,
                                cache=False,
                                download_cache=False)
    return query, paths


with ThreadPoolExecutor(max_workers=n_download_workers) as executor:
    futures = {executor.submit(_download_images_for_query, q): q
               for q in queries_to_download}
    for future in tqdm(as_completed(futures), total=len(futures)):
        query, paths = future.result()
        scientific_name_to_paths[query] = paths
        print('Downloaded {} images for {}'.format(len(paths),query))


#%% Rename .jpeg to .jpg

from megadetector.utils import path_utils
all_images = path_utils.recursive_file_list(image_base,False)

for fn in tqdm(all_images):
    if fn.lower().endswith('.jpeg'):
        new_fn = fn[0:-5] + '.jpg'
        os.rename(fn, new_fn)


#%% Choose representative images for each scientific name

# Specifically, sort by size, and take the largest unique sizes. Very small files tend
# to be bogus thumbnails, etc.

max_images_per_query = 4
scientific_name_to_preferred_images = {}

# s = list(scientific_name_to_paths.keys())[0]
for s in list(df.scientific_name):

    if not isinstance(s,str):
        continue

    query = s.replace(' ','+')

    if query in remapped_queries:
        query = remapped_queries[query]

    query_folder = os.path.join(image_base,query)
    assert os.path.isdir(query_folder)
    image_files = os.listdir(query_folder)
    image_fullpaths = [os.path.join(query_folder,fn) for fn in image_files]
    sizes = [os.path.getsize(p) for p in image_fullpaths]
    path_to_size = {}
    for i_fp,fp in enumerate(image_fullpaths):
        path_to_size[fp] = sizes[i_fp]
    paths_by_size = [x for _, x in sorted(zip(sizes, image_fullpaths),reverse=True)]

    # Be suspicious of duplicate sizes, those are typically duplicated images
    b_duplicate_sizes = [False] * len(paths_by_size)

    for i_path,p in enumerate(paths_by_size):
        if i_path == len(paths_by_size) - 1:
            continue
        if path_to_size[p] == path_to_size[paths_by_size[i_path+1]]:
            b_duplicate_sizes[i_path] = True

    paths_by_size_non_dup = [i for (i, v) in zip(paths_by_size, b_duplicate_sizes) if not v]

    preferred_paths = paths_by_size_non_dup[:max_images_per_query]
    scientific_name_to_preferred_images[s] = preferred_paths

# ...for each scientific name


#%% Delete unused images

used_images = []
for images in scientific_name_to_preferred_images.values():
    used_images.extend(images)

print('Using a total of {} images'.format(len(used_images)))
used_images_set = set(used_images)

from megadetector.utils import path_utils
all_images = path_utils.recursive_file_list(image_base,False)

unused_images = []
for fn in all_images:
    if fn not in used_images_set:
        unused_images.append(fn)

print('{} of {} files unused (diff {})'.format(len(unused_images),len(all_images),
                                               len(all_images) - len(unused_images)))

for fn in tqdm(unused_images):
    os.remove(fn)


#%% Produce HTML preview

with open(html_output_file, 'w', encoding='utf-8') as f:

    f.write('<html><head></head><body>\n')

    names = scientific_name_to_preferred_images.keys()
    names = sorted(names)

    f.write('<p class="speciesinfo_p" style="font-weight:bold;font-size:130%">'
            'dataset_name: <b><u>category</u></b> mapped to taxonomy_level scientific_name (taxonomic_common_name) (manual_common_name)</p>\n'
            '</p>')

    # i_row = 2; row = df.iloc[i_row]
    for i_row, row in tqdm(df.iterrows(), total=len(df)):

        s = row['scientific_name']

        taxonomy_string = row['taxonomy_string']
        if isinstance(taxonomy_string,str):
            taxonomic_match = eval(taxonomy_string)
            matched_entity = taxonomic_match[0]
            assert len(matched_entity) == 4
            common_names = matched_entity[3]
            if len(common_names) == 1:
                common_name_string = common_names[0]
            else:
                common_name_string = str(common_names)
        else:
            common_name_string = ''

        f.write('<p class="speciesinfo_p" style="font-weight:bold;font-size:130%">')

        if isinstance(row.scientific_name,str):
            output_string = '{}: <b><u>{}</u></b> mapped to {} {} ({}) ({})</p>\n'.format(
                row.dataset_name, row.query,
                row.taxonomy_level, row.scientific_name, common_name_string,
                row.common_name)
            f.write(output_string)
        else:
            f.write('{}: <b><u>{}</u></b> unmapped'.format(row.dataset_name,row.query))

        if s is None or s not in names:
            f.write('<p class="content_p">no images available</p>')
        else:
            image_paths = scientific_name_to_preferred_images[s]
            basedir = os.path.dirname(html_output_file)
            relative_paths = [os.path.relpath(p,basedir) for p in image_paths]
            image_paths = [s.replace('\\','/') for s in relative_paths]
            n_images = len(image_paths)
            # image_paths = [os.path.relpath(p, output_base) for p in image_paths]
            image_width_percent = round(100 / n_images)
            f.write('<table class="image_table"><tr>\n')
            for image_path in image_paths:
                f.write('<td style="vertical-align:top;" width="{}%">'
                        '<img src="{}" style="display:block; width:100%; vertical-align:top; height:auto;">'
                        '</td>\n'.format(image_width_percent, image_path))
            f.write('</tr></table>\n')

    # ...for each row

    f.write('</body></html>\n')


#%% Open HTML preview

from megadetector.utils.path_utils import open_file
open_file(html_output_file)



"""

labelme_to_coco.py

Converts a folder of labelme-formatted .json files to COCO.

"""

#%% Constants and imports

import ast
import os
import sys
import json
import argparse

from multiprocessing.pool import Pool, ThreadPool
from functools import partial
from tqdm import tqdm

from megadetector.utils import path_utils
from megadetector.visualization.visualization_utils import open_image


#%% Support functions

def _add_category(category_name,category_name_to_id,candidate_category_id=0):
    """
    Adds the category [category_name] to the dict [category_name_to_id], by default
    using the next available integer index.
    """

    if category_name in category_name_to_id:
        return category_name_to_id[category_name]
    while candidate_category_id in category_name_to_id.values():
        candidate_category_id += 1
    category_name_to_id[category_name] = candidate_category_id
    return candidate_category_id


def _process_labelme_file(image_fn_relative,
                          input_folder,
                          use_folders_as_labels,
                          no_json_handling,
                          validate_image_sizes,
                          category_name_to_id,
                          allow_new_categories=True):
    """
    Internal function for processing each image; this support function facilitates parallelization.
    """

    result = {}
    result['im'] = None
    result['annotations_this_image'] = None
    result['status'] = None

    image_fn_abs = os.path.join(input_folder,image_fn_relative)
    json_fn_abs = os.path.splitext(image_fn_abs)[0] + '.json'

    im = {}
    im['id'] = image_fn_relative
    im['file_name'] = image_fn_relative

    # If there's no .json file for this image...
    if not os.path.isfile(json_fn_abs):

        # Either skip it...
        if no_json_handling == 'skip':
            print('Skipping image {} (no .json file)'.format(image_fn_relative))
            result['status'] = 'skipped (no .json file)'
            return result

        # ...or error
        elif no_json_handling == 'error':
            raise ValueError('Image file {} has no corresponding .json file'.format(
                image_fn_relative))

        # ...or treat it as empty.
        elif no_json_handling == 'empty':
            try:
                pil_im = open_image(image_fn_abs)
            except Exception:
                print('Warning: error opening image {}, skipping'.format(image_fn_abs))
                result['status'] = 'image load error'
                return result
            im['width'] = pil_im.width
            im['height'] = pil_im.height

            # Just in case we need to differentiate between "no .json file" and "a .json file with no annotations"
            im['no_labelme_json'] = True
            shapes = []
        else:
            raise ValueError('Unrecognized specifier {} for handling images with no .json files'.format(
                no_json_handling))

    # If we found a .json file for this image...
    else:

        # Read the .json file
        with open(json_fn_abs,'r') as f:
            labelme_data = json.load(f)
        im['width'] = labelme_data['imageWidth']
        im['height'] = labelme_data['imageHeight']

        if validate_image_sizes:
            try:
                pil_im = open_image(image_fn_abs)
            except Exception:
                print('Warning: error opening image {} for size validation, skipping'.format(image_fn_abs))
                result['status'] = 'skipped (size validation error)'
                return result
            if not (im['width'] == pil_im.width and im['height'] == pil_im.height):
                print('Warning: image size validation error for file {}'.format(image_fn_relative))
                im['width'] = pil_im.width
                im['height'] = pil_im.height
                im['labelme_width'] = labelme_data['imageWidth']
                im['labelme_height'] = labelme_data['imageHeight']

        shapes = labelme_data['shapes']

        if ('flags' in labelme_data) and (len(labelme_data['flags']) > 0):
            im['flags'] = labelme_data['flags']

    annotations_this_image = []

    if len(shapes) == 0:

        if allow_new_categories:
            category_id = _add_category('empty',category_name_to_id)
        else:
            assert 'empty' in category_name_to_id
            category_id = category_name_to_id['empty']

        ann = {}
        ann['id'] = im['id'] + '_ann'
        ann['image_id'] = im['id']
        ann['category_id'] = category_id
        ann['sequence_level_annotation'] = False
        annotations_this_image.append(ann)

    else:

        for i_shape,shape in enumerate(shapes):

            if shape['shape_type'] not in ('rectangle','polygon'):
                print('Skipping an annotation of type {} in {}'.format(
                    shape['shape_type'],image_fn_relative))
                continue

            ann = {}
            annotations_this_image.append(ann)

            if use_folders_as_labels:
                category_name = os.path.basename(os.path.dirname(image_fn_abs))
            else:

                label = shape['label'].strip()
                assert len(label) > 0, 'Illegal label for file {}'.format(image_fn_relative)

                category_name = label

                # When multiple fields are available, labelme populates a dictionary (or at least
                # I've seen this happen once)
                if label[0] == '{' and label[-1] == '}' and 'name' in label:

                    try:
                        label_dict = ast.literal_eval(label)
                        category_name = label_dict['name']
                        # Copy all other fields directly to the output annotation
                        for k in label_dict:
                            if k != 'name':
                                ann[k] = label_dict[k]
                    except Exception:
                        print('Warning: failed to parse what looked like a dictionary for {}'.format(image_fn_relative))

                        # This is a no-op, but including it for clarity
                        category_name = label

                # ...if we might have to parse a dict for this label

            if allow_new_categories:
                category_id = _add_category(category_name,category_name_to_id)
            else:
                assert category_name in category_name_to_id
                category_id = category_name_to_id[category_name]

            points = shape['points']

            # Populate non-geometry fields
            ann['id'] = ann['id'] = im['id'] + '_ann_' + str(i_shape).zfill(3)
            ann['image_id'] = im['id']
            ann['category_id'] = category_id
            ann['sequence_level_annotation'] = False

            # Populate geometry
            if shape['shape_type'] == 'rectangle':

                if len(points) != 2:
                    print('Warning: illegal rectangle with {} points for {}'.format(
                        len(points),image_fn_relative))
                    continue

                p0 = points[0]
                p1 = points[1]
                x0 = min(p0[0],p1[0])
                x1 = max(p0[0],p1[0])
                y0 = min(p0[1],p1[1])
                y1 = max(p0[1],p1[1])

                bbox = [x0,y0,abs(x1-x0),abs(y1-y0)]
                ann['bbox'] = bbox

            else:

                assert shape['shape_type'] == 'polygon'

                if len(points) < 3:
                    print('Warning: illegal polygon with {} points for {}'.format(
                        len(points),image_fn_relative))
                    continue

                # COCO segmentation is a list of polygons, each a flat list of coordinates
                segmentation = []
                for pt in points:
                    segmentation.append(pt[0])
                    segmentation.append(pt[1])

                # Compute the axis-aligned bounding box
                x_coords = [pt[0] for pt in points]
                y_coords = [pt[1] for pt in points]
                x0 = min(x_coords)
                y0 = min(y_coords)
                bbox_w = max(x_coords) - x0
                bbox_h = max(y_coords) - y0

                bbox = [x0,y0,bbox_w,bbox_h]
                ann['bbox'] = bbox
                ann['segmentation'] = [segmentation]

            # ...if this is a rectangle/polygon

        # ...for each shape

    # ...if we do/don't have shapes for this image

    result['im'] = im
    result['annotations_this_image'] = annotations_this_image

    return result

# ...def _process_labelme_file(...)


#%% Main function

def labelme_to_coco(input_folder,
                    output_file=None,
                    category_id_to_category_name=None,
                    empty_category_name='empty',
                    empty_category_id=None,
                    info_struct=None,
                    relative_paths_to_include=None,
                    relative_paths_to_exclude=None,
                    use_folders_as_labels=False,
                    recursive=True,
                    no_json_handling='skip',
                    validate_image_sizes=True,
                    max_workers=1,
                    use_threads=True):
    """
    Finds all images in [input_folder] that have corresponding .json files, and converts
    to a COCO .json file.

    Currently supports bounding box and polygon annotations, as well as image-level flags
    (i.e., does not support point annotations).  Polygon annotations produce COCO annotations
    with a "segmentation" field and an axis-aligned bounding box.

    Labelme's image-level flags don't quite fit the COCO annotations format, so they are attached
    to image objects, rather than annotation objects.

    If output_file is None, just returns the resulting dict, does not write to file.

    if use_folders_as_labels is False (default), the output labels come from the labelme
    .json files.  If use_folders_as_labels is True, the lowest-level folder name containing
    each .json file will determine the output label.  E.g., if use_folders_as_labels is True,
    and the folder contains:

    images/train/lion/image0001.json

    ...all boxes in image0001.json will be given the label "lion", regardless of the labels in the
    file.  Empty images in the "lion" folder will still be given the label "empty" (or
    [empty_category_name]).

    Args:
        input_folder (str): input folder to search for images and Labelme .json files
        output_file (str, optional): output file to which we should write COCO-formatted data; if None
            this function just returns the COCO-formatted dict
        category_id_to_category_name (dict, optional): dict mapping category IDs to category names;
            really used to map Labelme category names to COCO category IDs.  IDs will be auto-generated
            if this is None.
        empty_category_name (str, optional): if images are present without boxes, the category name
            we should use for whole-image (and not-very-COCO-like) empty categories.
        empty_category_id (int, optional): category ID to use for the not-very-COCO-like "empty" category;
            also see the no_json_handling parameter.
        info_struct (dict, optional): dict to stash in the "info" field of the resulting COCO dict
        relative_paths_to_include (list, optional): allowlist of relative paths to include in the COCO
            dict; there's no reason to specify this along with relative_paths_to_exclude.
        relative_paths_to_exclude (list, optional): blocklist of relative paths to exclude from the COCO
            dict; there's no reason to specify this along with relative_paths_to_include.
        use_folders_as_labels (bool, optional): if this is True, class names will be pulled from folder names,
            useful if you have images like a/b/cat/image001.jpg, a/b/dog/image002.jpg, etc.
        recursive (bool, optional): whether to recurse into [input_folder]
        no_json_handling (str, optional): how to deal with image files that have no corresponding .json files,
            can be:

                - 'skip': ignore image files with no corresponding .json files
                - 'empty': treat image files with no corresponding .json files as empty
                - 'error': throw an error when an image file has no corresponding .json file
        validate_image_sizes (bool, optional): whether to load images to verify that the sizes specified
            in the labelme files are correct
        max_workers (int, optional): number of workers to use for parallelization, set to <=1 to disable
            parallelization
        use_threads (bool, optional): whether to use threads (True) or processes (False) for parallelization,
            not relevant if max_workers <= 1

    Returns:
        dict: a COCO-formatted dictionary, identical to what's written to [output_file] if [output_file] is not None.
    """

    if max_workers > 1:
        assert category_id_to_category_name is not None, \
            'When parallelizing labelme --> COCO conversion, you must supply a category mapping'

    if category_id_to_category_name is None:
        category_name_to_id = {}
    else:
        category_name_to_id = {v: k for k, v in category_id_to_category_name.items()}
    for category_name in category_name_to_id:
        try:
            category_name_to_id[category_name] = int(category_name_to_id[category_name])
        except ValueError:
            raise ValueError('Category IDs must be ints or string-formatted ints')

    # If the user supplied an explicit empty category ID, and the empty category
    # name is already in category_name_to_id, make sure they match.
    if empty_category_id is not None:
        if empty_category_name in category_name_to_id:
            assert category_name_to_id[empty_category_name] == empty_category_id, \
                'Ambiguous empty category specification'
        if empty_category_id in category_id_to_category_name:
            assert category_id_to_category_name[empty_category_id] == empty_category_name, \
                'Ambiguous empty category specification'
    else:
        if empty_category_name in category_name_to_id:
            empty_category_id = category_name_to_id[empty_category_name]

    del category_id_to_category_name

    # Enumerate images
    print('Enumerating images in {}'.format(input_folder))
    image_filenames_relative = path_utils.find_images(input_folder,
                                                      recursive=recursive,
                                                      return_relative_paths=True,
                                                      convert_slashes=True)

    # Remove any images we're supposed to skip
    if (relative_paths_to_include is not None) or (relative_paths_to_exclude is not None):

        image_filenames_relative_to_process = []
        for image_fn_relative in image_filenames_relative:
            if relative_paths_to_include is not None and image_fn_relative not in relative_paths_to_include:
                continue
            if relative_paths_to_exclude is not None and image_fn_relative in relative_paths_to_exclude:
                continue
            image_filenames_relative_to_process.append(image_fn_relative)
        print('Processing {} of {} images'.format(
            len(image_filenames_relative_to_process),
            len(image_filenames_relative)))
        image_filenames_relative = image_filenames_relative_to_process

    # If the user supplied a category ID to use for empty images...
    if empty_category_id is not None:
        try:
            empty_category_id = int(empty_category_id)
        except ValueError:
            raise ValueError('Category IDs must be ints or string-formatted ints')

    if empty_category_id is None:
        empty_category_id = _add_category(empty_category_name,category_name_to_id)

    print('Processing annotations...')

    if max_workers <= 1:

        image_results = []
        for image_fn_relative in tqdm(image_filenames_relative):

            result = _process_labelme_file(image_fn_relative,input_folder,use_folders_as_labels,
                                      no_json_handling,validate_image_sizes,
                                      category_name_to_id,allow_new_categories=True)
            image_results.append(result)

    else:

        n_workers = min(max_workers,len(image_filenames_relative))
        assert category_name_to_id is not None

        pool = None
        try:
            if use_threads:
                pool = ThreadPool(n_workers)
            else:
                pool = Pool(n_workers)

            image_results = list(tqdm(pool.imap(
                partial(_process_labelme_file,
                    input_folder=input_folder,
                    use_folders_as_labels=use_folders_as_labels,
                    no_json_handling=no_json_handling,
                    validate_image_sizes=validate_image_sizes,
                    category_name_to_id=category_name_to_id,
                    allow_new_categories=False
                    ),image_filenames_relative), total=len(image_filenames_relative)))
        finally:
            if pool is not None:
                pool.close()
                pool.join()
                print('Pool closed and joined for labelme file processing')

    images = []
    annotations = []

    # Flatten the lists of images and annotations
    for result in image_results:
        im = result['im']
        annotations_this_image = result['annotations_this_image']

        if im is None:
            assert annotations_this_image is None
        else:
            images.append(im)
            annotations.extend(annotations_this_image)

    output_dict = {}
    output_dict['images'] = images
    output_dict['annotations'] = annotations

    if info_struct is None:
        info_struct = {}
    if 'description' not in info_struct:
        info_struct['description'] = \
            'Converted to COCO from labelme annotations in folder {}'.format(input_folder)
    if 'version' not in info_struct:
        info_struct['version'] = 1.0

    output_dict['info'] = info_struct
    categories = []
    for category_name in category_name_to_id:
        categories.append({'name':category_name,'id':category_name_to_id[category_name]})
    output_dict['categories'] = categories

    if output_file is not None:
        with open(output_file,'w') as f:
            json.dump(output_dict,f,indent=1)

    return output_dict

# ...def labelme_to_coco()


def find_empty_labelme_files(input_folder,recursive=True):
    """
    Returns a list of all image files in in [input_folder] associated with .json files that have
    no boxes in them.  Also returns a list of images with no associated .json files.  Specifically,
    returns a dict:

    .. code-block: none

        {
            'images_with_empty_json_files':[list],
            'images_with_no_json_files':[list],
            'images_with_non_empty_json_files':[list]
        }

    Args:
        input_folder (str): the folder to search for empty (i.e., box-less) Labelme .json files
        recursive (bool, optional): whether to recurse into [input_folder]

    Returns:
        dict: a dict with fields:
            - images_with_empty_json_files: a list of all image files in [input_folder] associated with
              .json files that have no boxes in them
            - images_with_no_json_files: a list of images in [input_folder] with no associated .json files
            - images_with_non_empty_json_files: a list of images in [input_folder] associated with .json
              files that have at least one box
    """

    image_filenames_relative = path_utils.find_images(input_folder,
                                                      recursive=recursive,
                                                      return_relative_paths=True)

    images_with_empty_json_files = []
    images_with_no_json_files = []
    images_with_non_empty_json_files = []

    # fn_relative = image_filenames_relative[0]
    for fn_relative in image_filenames_relative:

        image_fn_abs = os.path.join(input_folder,fn_relative)
        json_fn_abs = os.path.splitext(image_fn_abs)[0] + '.json'

        if not os.path.isfile(json_fn_abs):
            images_with_no_json_files.append(fn_relative)
            continue

        else:
           # Read the .json file
           with open(json_fn_abs,'r') as f:
               labelme_data = json.load(f)
           shapes = labelme_data['shapes']
           if len(shapes) == 0:
               images_with_empty_json_files.append(fn_relative)
           else:
               images_with_non_empty_json_files.append(fn_relative)

    # ...for every image

    return {'images_with_empty_json_files':images_with_empty_json_files,
            'images_with_no_json_files':images_with_no_json_files,
            'images_with_non_empty_json_files':images_with_non_empty_json_files}

# ...def find_empty_labelme_files(...)


#%% Interactive driver

if False:

    pass

    #%% Options

    empty_category_name = 'empty'
    empty_category_id = None
    category_id_to_category_name = None
    info_struct = None

    input_folder = os.path.expanduser('~/data/md-test')
    output_file = os.path.expanduser('~/data/md-test-labelme-to-coco.json')


    #%% Programmatic execution

    output_dict = labelme_to_coco(input_folder,output_file,
                                  category_id_to_category_name=category_id_to_category_name,
                                  empty_category_name=empty_category_name,
                                  empty_category_id=empty_category_id,
                                  info_struct=None,
                                  use_folders_as_labels=False,
                                  validate_image_sizes=False,
                                  no_json_handling='empty')


    #%% Validate

    from megadetector.data_management.databases import integrity_check_json_db

    options = integrity_check_json_db.IntegrityCheckOptions()

    options.baseDir = input_folder
    options.bCheckImageSizes = True
    options.bCheckImageExistence = True
    options.bFindUnusedImages = True
    options.bRequireLocation = False

    sorted_categories, _, error_info = integrity_check_json_db.integrity_check_json_db(output_file,options)


    #%% Preview

    from megadetector.visualization import visualize_db
    options = visualize_db.DbVizOptions()
    options.parallelize_rendering = True
    options.viz_size = (900, -1)
    options.num_to_visualize = 5000

    html_file,_ = visualize_db.visualize_db(output_file,os.path.expanduser('~/tmp/labelme_to_coco_preview'),
                                input_folder,options)


    from megadetector.utils import path_utils # noqa
    path_utils.open_file(html_file)


    #%% Prepare command line

    s = 'python labelme_to_coco.py {} {}'.format(input_folder,output_file)
    print(s)
    import clipboard; clipboard.copy(s)


#%% Command-line driver

def main(): # noqa

    parser = argparse.ArgumentParser(
        description='Convert labelme-formatted data to COCO')

    parser.add_argument(
        'input_folder',
        type=str,
        help='Path to images and .json annotation files')

    parser.add_argument(
        'output_file',
        type=str,
        help='Output filename (.json)')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    labelme_to_coco(args.input_folder,args.output_file)

if __name__ == '__main__':
    main()



"""

generate_csv_report.py

Generates a .csv report from a MD-formatted .json file with the following columns:

* filename
* datetime (if images or EXIF information is supplied)
* detection_category
* max_detection_confidence
* classification_category
* max_classification_confidence
* count

One row is generated per category pair per image.  For example, these would be unique rows:

image0001.jpg,animal,deer,4
image0001.jpg,animal,lion,4
image0001.jpg,animal,[none],4
image0001.jpg,person,[none],2

Images with no above-threshold detections will have a single row:

image0001.jpg,empty,[none],-1

Images with processing errors will have a single row:

image0001.jpg,error,error_string,-1

"""

#%% Constants and imports

import os
import json
import tempfile
import sys
import argparse
import uuid

import pandas as pd

from copy import deepcopy

from megadetector.utils.wi_taxonomy_utils import load_md_or_speciesnet_file
from megadetector.utils.ct_utils import get_max_conf
from megadetector.utils.ct_utils import is_list_sorted
from megadetector.detection.run_detector import \
    get_typical_confidence_threshold_from_results
from megadetector.data_management.read_exif import \
    read_exif_from_folder, ReadExifOptions, minimal_exif_tags

default_classification_threshold = 0.3
unknown_datetime_tag = ''


#%% Functions

def generate_csv_report(md_results_file,
                        output_file=None,
                        datetime_source=None,
                        folder_level_columns=None,
                        detection_confidence_threshold=None,
                        classification_confidence_threshold=None,
                        verbose=True):
    """
    Generates a .csv report from a MD-formatted .json file

    Args:
        md_results_file (str): MD results .json file for which we should generate a report
        output_file (str, optional): .csv file to write; if this is None, we'll use md_results_file.csv
        datetime_source (str, optional): if datetime information is required, this should point to
            a folder of images, a MD results .json file (can be the same as the input file), or
            an exif_info.json file created with read_exif().
        folder_level_columns (list of int, optional): list of folder levels (where zero is the top-level
            folder in a path name) for which we should create separate columns.  Should be zero-indexed ints,
            or a comma-delimited list of zero-indexed int-strings.
        detection_confidence_threshold (float, optional): detections below this confidence threshold will not
            be included in the output data.  Defaults to the recommended value based on the .json file.
        classification_confidence_threshold (float, optional): classifications below this confidence threshold will
            not be included in the output data (i.e., detections will be considered "animal").
        verbose (bool, optional): enable debug output, including the progress bar,

    Returns:
        str: the output .csv filename
    """

    ##%% Load results file

    results = load_md_or_speciesnet_file(md_results_file)

    print('Loaded results for {} images'.format(len(results['images'])))

    detection_category_id_to_name = results['detection_categories']
    classification_category_id_to_name = None
    if 'classification_categories' in results:
        classification_category_id_to_name = results['classification_categories']

    if output_file is None:
        output_file = md_results_file + '.csv'

    ##%% Read datetime information if necessary

    filename_to_datetime_string = None

    if datetime_source is not None:

        all_exif_results = None

        if os.path.isdir(datetime_source):

            # Read EXIF info from images
            read_exif_options = ReadExifOptions()
            read_exif_options.tags_to_include = minimal_exif_tags
            read_exif_options.byte_handling = 'delete'
            exif_cache_file = os.path.join(tempfile.gettempdir(),
                                           'md-exif-data',
                                           str(uuid.uuid1())+'.json')
            print('Reading EXIF datetime info from {}, writing to {}'.format(
                datetime_source,exif_cache_file))
            os.makedirs(os.path.dirname(exif_cache_file),exist_ok=True)

            all_exif_results = read_exif_from_folder(input_folder=datetime_source,
                                                     output_file=exif_cache_file,
                                                     options=read_exif_options,
                                                     recursive=True)

        else:

            assert os.path.isfile(datetime_source), \
                'datetime source {} is neither a folder nor a file'.format(datetime_source)

            # Is this the same file we've already read?

            # Load this, decide whether it's a MD file or an exif_info file
            with open(datetime_source,'r') as f:
                d = json.load(f)

            if isinstance(d,list):
                all_exif_results = d
            else:
                assert isinstance(d,dict), 'Unrecognized file format supplied as datetime source'
                assert 'images' in d,\
                    'The datetime source you provided doesn\'t look like a valid source .json file'
                all_exif_results = []
                found_datetime = False
                for im in d['images']:
                    exif_result = {'file_name':im['file']}
                    if 'datetime' in im:
                        found_datetime = True
                        exif_result['exif_tags'] = {'DateTimeOriginal':im['datetime']}
                    all_exif_results.append(exif_result)
                if not found_datetime:
                    print('Warning: a MD results file was supplied as the datetime source, but it does not appear '
                          'to contain datetime information.')

        # ...if datetime_source is a folder/file

        assert all_exif_results is not None

        filename_to_datetime_string = {}

        for exif_result in all_exif_results:

            datetime_string = unknown_datetime_tag
            if ('exif_tags' in exif_result) and \
               (exif_result['exif_tags'] is not None) and \
               ('DateTimeOriginal' in exif_result['exif_tags']):
                datetime_string = exif_result['exif_tags']['DateTimeOriginal']
                if datetime_string is None:
                    datetime_string = ''
                else:
                    assert isinstance(datetime_string,str), 'Unrecognized datetime format'
            filename_to_datetime_string[exif_result['file_name']] = datetime_string

        # ...for each exif result

        image_files = [im['file'] for im in results['images']]
        image_files_set = set(image_files)

        files_in_exif_but_not_in_results = []
        files_in_results_but_not_in_exif = []
        files_with_no_datetime_info = []

        for fn in filename_to_datetime_string:
            dts = filename_to_datetime_string[fn]
            if (dts is None) or (dts == unknown_datetime_tag) or (len(dts) == 0):
                files_with_no_datetime_info.append(fn)
            if fn not in image_files_set:
                files_in_exif_but_not_in_results.append(fn)

        for fn in image_files_set:
            if fn not in filename_to_datetime_string:
                files_in_results_but_not_in_exif.append(fn)

        print('{} files (of {}) in EXIF info not found in MD results'.format(
            len(files_in_exif_but_not_in_results),len(filename_to_datetime_string)
        ))

        print('{} files (of {}) in MD results not found in MD EXIF info'.format(
            len(files_in_results_but_not_in_exif),len(image_files_set)
        ))

        print('Failed to read datetime information for {} files (of {}) in EXIF info'.format(
            len(files_with_no_datetime_info),len(filename_to_datetime_string)
        ))

    # ...if we need to deal with datetimes


    ##%% Parse folder level column specifier

    if folder_level_columns is not None:

        if isinstance(folder_level_columns,str):
            tokens = folder_level_columns.split(',')
            folder_level_columns = [int(s) for s in tokens]
        for folder_level in folder_level_columns:
            if (not isinstance(folder_level,int)) or (folder_level < 0):
                raise ValueError('Illegal folder level specifier {}'.format(
                    str(folder_level_columns)))


    ##%% Fill in default thresholds

    if classification_confidence_threshold is None:
        classification_confidence_threshold = default_classification_threshold
    if detection_confidence_threshold is None:
        detection_confidence_threshold = \
            get_typical_confidence_threshold_from_results(results)

    assert detection_confidence_threshold is not None


    ##%% Fill in output records

    output_records = []

    # For each image
    #
    # im = results['images'][0]
    for im in results['images']:

        """
        * filename
        * datetime (if images or EXIF information is supplied)
        * detection_category
        * max_detection_confidence
        * classification_category
        * max_classification_confidence
        * count
        """

        base_record = {}

        base_record['filename'] = im['file'].replace('\\','/')

        # Datetime (if necessary)
        datetime_string = ''
        if filename_to_datetime_string is not None:
            if im['file'] in filename_to_datetime_string:
                datetime_string = filename_to_datetime_string[im['file']]
        base_record['datetime'] = datetime_string

        for s in ['detection_category','max_detection_confidence',
                  'classification_category','max_classification_confidence',
                  'count']:
            base_record[s] = ''

        # Folder level columns
        tokens = im['file'].split('/')

        if folder_level_columns is not None:

            for folder_level in folder_level_columns:
                folder_level_column_name = 'folder_level_' + str(folder_level).zfill(2)
                if folder_level >= len(tokens):
                    folder_level_value = ''
                else:
                    folder_level_value = tokens[folder_level]
                base_record[folder_level_column_name] = folder_level_value

        records_this_image = []

        # Create one output row if this image failed
        if 'failure' in im and im['failure'] is not None and len(im['failure']) > 0:

            record = deepcopy(base_record)
            record['detection_category'] = 'error'
            record['classification_category'] = im['failure']
            records_this_image.append(record)
            assert 'detections' not in im or im['detections'] is None

        else:

            assert 'detections' in im and im['detections'] is not None

            # Count above-threshold detections
            detections_above_threshold = []
            for det in im['detections']:
                if det['conf'] >= detection_confidence_threshold:
                    detections_above_threshold.append(det)
            max_detection_conf = get_max_conf(im)

            # Create one output row if this image is empty (i.e., has no
            # above-threshold detections)
            if len(detections_above_threshold) == 0:

                record = deepcopy(base_record)
                record['detection_category'] = 'empty'
                record['max_detection_confidence'] = max_detection_conf
                records_this_image.append(record)

            # ...if this image is empty

            else:

                # Maps a string of the form:
                #
                # detection_category:classification_category
                #
                # ...to a dict with fields ['max_detection_conf','max_classification_conf','count']
                category_info_string_to_record = {}

                for det in detections_above_threshold:

                    assert det['conf'] >= detection_confidence_threshold

                    detection_category_name = detection_category_id_to_name[det['category']]
                    detection_confidence = det['conf']
                    classification_category_name = ''
                    classification_confidence = 0.0

                    if ('classifications' in det) and (len(det['classifications']) > 0):

                        # Classifications should always be sorted by confidence.  Not
                        # technically required, but always true in practice.
                        assert is_list_sorted([c[1] for c in det['classifications']]), \
                            'This script does not yet support unsorted classifications'
                        assert classification_category_id_to_name is not None, \
                            'If classifications are present, category mappings should be present'

                        # Only use the first classification
                        classification = det['classifications'][0]
                        if classification[1] >= classification_confidence_threshold:
                            classification_category_name = \
                                classification_category_id_to_name[classification[0]]
                            classification_confidence = classification[1]

                    # ...if classifications are present

                    # E.g. "animal:rodent", or "vehicle:"
                    category_info_string = detection_category_name + ':' + classification_category_name

                    if category_info_string not in category_info_string_to_record:
                        category_info_string_to_record[category_info_string] = {
                            'max_detection_confidence':0.0,
                            'max_classification_confidence':0.0,
                            'count':0,
                            'detection_category':detection_category_name,
                            'classification_category':classification_category_name
                        }

                    record = category_info_string_to_record[category_info_string]
                    record['count'] += 1
                    if detection_confidence > record['max_detection_confidence']:
                        record['max_detection_confidence'] = detection_confidence
                    if classification_confidence > record['max_classification_confidence']:
                        record['max_classification_confidence'] = classification_confidence

                # ...for each detection

                for record_in in category_info_string_to_record.values():
                    assert record_in['count'] > 0
                    record_out = deepcopy(base_record)
                    for k in record_in.keys():
                        assert k in record_out.keys()
                        record_out[k] = record_in[k]
                    records_this_image.append(record_out)

            # ...is this empty/non-empty?

        # ...if this image failed/didn't fail

        # Add to [records]
        output_records.extend(records_this_image)

    # ...for each image

    # Make sure every record has the same columns

    if len(output_records) == 0:
        print('Warning: no output records generated')
    else:
        column_names = output_records[0].keys()
        for record in output_records:
            assert record.keys() == column_names

        # Create folder for output file if necessary
        output_dir = os.path.dirname(output_file)
        if len(output_dir) > 0:
            os.makedirs(output_dir, exist_ok=True)

        # Write to .csv
        df = pd.DataFrame(output_records)
        df.to_csv(output_file,header=True,index=False)

    # from megadetector.utils.path_utils import open_file; open_file(output_file)

# ...generate_csv_report(...)


# %%

#%% Interactive driver

if False:

    pass

    #%% Configure options

    r"""
    python run_detector_batch.py MDV5A "g:\temp\md-test-images"
    "g:\temp\md-test-images\md_results_with_datetime.json"
    --recursive --output_relative_filenames --include_image_timestamp --include_exif_data
    """

    md_results_file = 'g:/temp/csv-report-test/md-results.json'
    datetime_source = 'g:/temp/csv-report-test/exif_data.json'

    # datetime_source = 'g:/temp/md-test-images'
    # datetime_source = 'g:/temp/md-test-images/md_results_with_datetime.json'
    # md_results_file = 'g:/temp/md-test-images/md_results_with_datetime.json'
    # md_results_file = 'g:/temp/md-test-images/speciesnet_results_md_format.json'

    output_file = None
    folder_level_columns = [0,1,2,3]
    detection_confidence_threshold = None
    classification_confidence_threshold = None
    verbose = True


    #%% Programmatic execution

    generate_csv_report(md_results_file=md_results_file,
                        output_file=output_file,
                        datetime_source=datetime_source,
                        folder_level_columns=folder_level_columns,
                        detection_confidence_threshold=detection_confidence_threshold,
                        classification_confidence_threshold=classification_confidence_threshold,
                        verbose=verbose)


#%% Command-line driver

def main(): # noqa

    parser = argparse.ArgumentParser(
        description='Generates a .csv report from a MD-formatted .json file')

    parser.add_argument(
        'md_results_file',
        type=str,
        help='Path to MD results file (.json)')

    parser.add_argument(
        '--output_file',
        type=str,
        help='Output filename (.csv) (if omitted, will append .csv to the input file)')

    parser.add_argument(
        '--datetime_source',
        type=str,
        default=None,
        help='Image folder, exif_info.json file, or MD results file from which we should read datetime information'
        )

    parser.add_argument(
        '--folder_level_columns',
        type=str,
        default=None,
        help='Comma-separated list of zero-indexed folder levels that should become columns in the output file'
        )

    parser.add_argument(
        '--detection_confidence_threshold',
        type=float,
        default=None,
        help='Detection threshold (if omitted, chooses a reasonable default based on the .json file)'
        )

    parser.add_argument(
        '--classification_confidence_threshold',
        type=float,
        default=None,
        help='Classification threshold (default {})'.format(default_classification_threshold)
        )

    parser.add_argument(
        '--verbose',
        action='store_true',
        help='Enable additional debug output'
        )


    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    generate_csv_report(md_results_file=args.md_results_file,
                        output_file=args.output_file,
                        datetime_source=args.datetime_source,
                        folder_level_columns=args.folder_level_columns,
                        detection_confidence_threshold=args.detection_confidence_threshold,
                        classification_confidence_threshold=args.classification_confidence_threshold,
                        verbose=args.verbose)

if __name__ == '__main__':
    main()



"""

subset_json_db.py

Select a subset of images (and associated annotations) from a .json file in COCO
Camera Traps format based on a string query.

To subset .json files in the MegaDetector output format, see
subset_json_detector_output.py.

"""

#%% Constants and imports

import sys
import json
import argparse

from tqdm import tqdm
from copy import copy

from megadetector.utils import ct_utils
from megadetector.utils.ct_utils import sort_list_of_dicts_by_key


#%% Functions

def subset_json_db(input_json,
                   query,
                   output_json=None,
                   ignore_case=False,
                   remap_categories=True,
                   verbose=False):
    """
    Given a json file (or dictionary already loaded from a json file), produce a new
    database containing only the images whose filenames contain the string 'query',
    optionally writing that DB output to a new json file.

    Args:
        input_json (str): COCO Camera Traps .json file to load, or an already-loaded dict
        query (str or list): string to query for, only include images in the output whose filenames
            contain this string.  If this is a list, test for exact matches.
        output_json (str, optional): file to write the resulting .json file to
        ignore_case (bool, optional): whether to perform a case-insensitive search for [query]
        remap_categories (bool, optional): trim the category list to only the categores used
            in the subset
        verbose (bool, optional): enable additional debug output

    Returns:
        dict: CCT dictionary containing a subset of the images and annotations in the input dict
    """

    # Load the input file if necessary
    if isinstance(input_json,str):
        print('Loading input .json...')
        with open(input_json, 'r') as f:
            input_data = json.load(f)
    else:
        input_data = input_json

    # Find images matching the query
    images = []

    if isinstance(query,str):

        if ignore_case:
            query = query.lower()

        for im in tqdm(input_data['images']):
            fn = im['file_name']
            if ignore_case:
                fn = fn.lower()
            if query in fn:
                images.append(im)

    else:

        query = set(query)

        if ignore_case:
            query = set([s.lower() for s in query])

        for im in input_data['images']:
            fn = im['file_name']
            if ignore_case:
                fn = fn.lower()
            if fn in query:
                images.append(im)

    image_ids = set([im['id'] for im in images])

    # Find annotations referring to those images
    annotations = []

    for ann in input_data['annotations']:
        if ann['image_id'] in image_ids:
            annotations.append(ann)

    output_data = copy(input_data)
    output_data['images'] = images
    output_data['annotations'] = annotations

    # Remap categories if necessary
    if remap_categories:

        category_ids_used = set()
        for ann in annotations:
            category_ids_used.add(ann['category_id'])

        if verbose:
            print('Keeping {} of {} categories'.format(
                len(category_ids_used),len(input_data['categories'])))

        input_category_id_to_output_category_id = {}

        next_category_id = 0

        # Build mappings from old to new category IDs
        for input_category_id in category_ids_used:
            assert isinstance(input_category_id,int), \
                'Illegal category ID {}'.format(input_category_id)
            output_category_id = next_category_id
            next_category_id = next_category_id + 1
            input_category_id_to_output_category_id[input_category_id] = output_category_id

        # Modify the annotations
        for ann in annotations:
            assert ann['category_id'] in input_category_id_to_output_category_id
            ann['category_id'] = input_category_id_to_output_category_id[ann['category_id']]

        output_categories = []

        # Re-write the category table
        for cat in input_data['categories']:

            if cat['id'] in input_category_id_to_output_category_id:

                # There may be non-required fields, so don't just create an empty dict
                # and copy the name/id field, keep the original dict other than "id"
                output_category = copy(cat)
                output_category['id'] = input_category_id_to_output_category_id[cat['id']]
                output_categories.append(output_category)

        output_categories = sort_list_of_dicts_by_key(output_categories,'id')
        output_data['categories'] = output_categories

    # ...if we need to remap categories

    # Write the output file if requested
    if output_json is not None:
        if verbose:
            print('Writing output .json to {}'.format(output_json))
        ct_utils.write_json(output_json, output_data)

    if verbose:
        print('Keeping {} of {} images, {} of {} annotations'.format(
            len(output_data['images']),len(input_data['images']),
            len(output_data['annotations']),len(input_data['annotations'])))

    return output_data


#%% Interactive driver

if False:

    #%%

    input_json = r"e:\Statewide_wolf_container\idfg_20190409.json"
    output_json = r"e:\Statewide_wolf_container\idfg_20190409_clearcreek.json"
    query = 'clearcreek'
    ignore_case = True
    db = subset_json_db(input_json, query, output_json, ignore_case)


#%% Command-line driver

def main(): # noqa

    parser = argparse.ArgumentParser()
    parser.add_argument('input_json', type=str, help='Input file (a COCO Camera Traps .json file)')
    parser.add_argument('output_json', type=str, help='Output file')
    parser.add_argument('query', type=str, help='Filename query')
    parser.add_argument('--ignore_case', action='store_true')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    subset_json_db(args.input_json,args.query,args.output_json,args.ignore_case)

if __name__ == '__main__':
    main()



"""

load_api_results.py

DEPRECATED

As of 2023.12, this module is still used in postprocessing and RDE, but it's not recommended
for new code.

Loads the output of the batch processing API (json) into a Pandas dataframe.

Includes functions to read/write the (very very old) .csv results format.

"""

#%% Imports

import json
import os

from typing import Optional
from collections.abc import Mapping

import pandas as pd

from megadetector.utils.ct_utils import get_max_conf
from megadetector.utils.ct_utils import write_json
from megadetector.utils.wi_taxonomy_utils import load_md_or_speciesnet_file


#%% Functions for loading .json results into a Pandas DataFrame, and writing back to .json

def load_api_results(api_output_path: str, normalize_paths: bool = True,
                     filename_replacements: Optional[Mapping[str, str]] = None,
                     force_forward_slashes: bool = True
                     ) -> tuple[pd.DataFrame, dict]:
    r"""
    Loads json-formatted MegaDetector results to a Pandas DataFrame.

    Args:
        api_output_path (str): path to the output json file
        normalize_paths (bool, optional): whether to apply os.path.normpath to the 'file'
            field in each image entry in the output file
        filename_replacements (dict, optional): replace some path tokens to match local paths
            to the original file structure
        force_forward_slashes (bool, optional): whether to convert backslashes to forward
            slashes in filenames

    Returns:
        detection_results: pd.DataFrame, contains at least the columns ['file', 'detections','failure']
        other_fields: a dict containing fields in the results other than 'images'
    """

    print('Loading results from {}'.format(api_output_path))

    detection_results = load_md_or_speciesnet_file(api_output_path)

    # Validate that this is really a detector output file
    for s in ['info', 'detection_categories', 'images']:
        assert s in detection_results, 'Missing field {} in detection results'.format(s)

    # Fields in the output json other than 'images'
    other_fields = {}
    for k, v in detection_results.items():
        if k != 'images':
            other_fields[k] = v

    if normalize_paths:
        for image in detection_results['images']:
            image['file'] = os.path.normpath(image['file'])

    if force_forward_slashes:
        for image in detection_results['images']:
            image['file'] = image['file'].replace('\\','/')

    # Replace some path tokens to match local paths to original blob structure
    if filename_replacements is not None:
        for string_to_replace in filename_replacements.keys():
            replacement_string = filename_replacements[string_to_replace]
            for im in detection_results['images']:
                im['file'] = im['file'].replace(string_to_replace,replacement_string)

    print('Converting results to dataframe')

    # If this is a newer file that doesn't include maximum detection confidence values,
    # add them, because our unofficial internal dataframe format includes this.
    for im in detection_results['images']:
        if 'max_detection_conf' not in im:
            im['max_detection_conf'] = get_max_conf(im)

    # Pack the json output into a Pandas DataFrame
    detection_results = pd.DataFrame(detection_results['images'])

    print('Finished loading MegaDetector results for {} images from {}'.format(
            len(detection_results),api_output_path))

    return detection_results, other_fields


def write_api_results(detection_results_table, other_fields, out_path):
    """
    Writes a Pandas DataFrame to the MegaDetector .json format.

    Args:
        detection_results_table (DataFrame): data to write
        other_fields (dict): additional fields to include in the output .json
        out_path (str): output .json filename
    """

    print('Writing detection results to {}'.format(out_path))

    fields = other_fields

    images = detection_results_table.to_json(orient='records',
                                             double_precision=3)
    images = json.loads(images)
    for im in images:
        if 'failure' in im and im['failure'] is None:
            del im['failure']
    fields['images'] = images

    # Convert the 'version' field back to a string as per format convention
    try:
        version = other_fields['info']['format_version']
        if not isinstance(version,str):
            other_fields['info']['format_version'] = str(version)
    except Exception:
        print('Warning: error determining format version')
        pass

    # Remove 'max_detection_conf' as per newer file convention (format >= v1.3)
    try:
        version = other_fields['info']['format_version']
        version = float(version)
        if version >= 1.3:
            for im in images:
                if 'max_detection_conf' in im:
                    del im['max_detection_conf']
    except Exception:
        print('Warning: error removing max_detection_conf from output')
        pass

    write_json(out_path,fields)

    print('Finished writing detection results to {}'.format(out_path))


def load_api_results_csv(filename, normalize_paths=True, filename_replacements=None, nrows=None):
    """
    [DEPRECATED]

    Loads .csv-formatted MegaDetector results to a pandas table

    Args:
        filename (str): path to the csv file to read
        normalize_paths (bool, optional): whether to apply os.path.normpath to the 'file'
            field in each image entry in the output file
        filename_replacements (dict, optional): replace some path tokens to match local paths
            to the original file structure
        nrows (int, optional): read only the first N rows of [filename]
    """

    if filename_replacements is None:
        filename_replacements = {}

    print('Loading MegaDetector results from {}'.format(filename))

    detection_results = pd.read_csv(filename,nrows=nrows)

    print('De-serializing MegaDetector results from {}'.format(filename))

    # Confirm that this is really a detector output file
    for s in ['image_path','max_confidence','detections']:
        assert s in detection_results.columns

    # Normalize paths to simplify comparisons later
    if normalize_paths:
        detection_results['image_path'] = detection_results['image_path'].apply(os.path.normpath)

    # De-serialize detections
    detection_results['detections'] = detection_results['detections'].apply(json.loads)

    # Optionally replace some path tokens to match local paths to the original blob structure
    # string_to_replace = list(options.detector_output_filename_replacements.keys())[0]
    for string_to_replace in filename_replacements:

        replacement_string = filename_replacements[string_to_replace]

        # i_row = 0
        for i_row in range(0,len(detection_results)):
            row = detection_results.iloc[i_row]
            fn = row['image_path']
            fn = fn.replace(string_to_replace,replacement_string)
            detection_results.at[i_row,'image_path'] = fn

    print('Finished loading and de-serializing MD results for {} images from {}'.format(
        len(detection_results),filename))

    return detection_results


def write_api_results_csv(detection_results, filename):
    """
    [DEPRECATED]

    Writes a Pandas table to csv in a way that's compatible with the .csv output
    format.  Currently just a wrapper around to_csv that forces output writing
    to go through a common code path.

    Args:
        detection_results (DataFrame): dataframe to write to [filename]
        filename (str): .csv filename to write
    """

    print('Writing detection results to {}'.format(filename))

    output_dir = os.path.dirname(filename)
    if len(output_dir) > 0:
        os.makedirs(output_dir, exist_ok=True)

    detection_results.to_csv(filename, index=False)

    print('Finished writing detection results to {}'.format(filename))



"""

retrieve_sample_image.py

Downloader that retrieves images from Bing  images, used for verifying taxonomy
lookups and looking for egregious mismappings (e.g., "snake" being mapped to a fish called
"snake").

Simple wrapper around simple_image_download, but I've had to swap in and out the underlying
downloader a few times, so I use an extra layer of indirection.

"""

#%% Imports and environment

import os

from megadetector.taxonomy_mapping import simple_image_download

default_output_folder = os.path.expanduser('~/tmp/image-download-test')

image_downloader = simple_image_download.Downloader()


#%% Main entry point

def download_images(query,
                    output_directory=default_output_folder,
                    limit=100,
                    verbose=False):

    query = query.replace(' ','+')

    image_downloader.directory = output_directory
    paths = image_downloader.download(query,
                                      limit=limit,
                                      verbose=verbose,
                                      cache=False,
                                      download_cache=False)
    return paths


#%% Test driver

if False:

    pass

    #%%

    paths = download_images(query='redunca',
                            output_directory=output_folder,
                            limit=20,
                            verbose=True)
